"""Module for interacting with the bio.tools API."""

import uuid
from collections.abc import Callable
from typing import TYPE_CHECKING

import requests
from langchain.chains.openai_functions import create_structured_output_runnable
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.pydantic_v1 import BaseModel, Field

if TYPE_CHECKING:
    from biochatter.llm_connect import Conversation

from biochatter.api_agent.base.agent_abc import (
    BaseFetcher,
    BaseInterpreter,
    BaseQueryBuilder,
)

BIOTOOLS_QUERY_PROMPT = """
You are a world class algorithm for creating queries in structured formats. Your task is to use the web API of bio.tools
to answer questions about bioinformatics tools and their properties.

You have to extract the appropriate information out of the examples:
1. To list information about the tools, use the endpoint https://bio.tools/api/t/ with parameters like name,
description, homepage, etc.

Use these formats to generate queries based on the question provided. Below is more information about the bio.tools API:

Base URL

https://bio.tools/api/

Endpoints and Parameters

1. List tools

GET /t/

==================  ============================================================================================
Parameter           Search behaviour
==================  ============================================================================================
biotoolsID          Search for bio.tools tool ID (usually quoted - to get exact match)

                    `biotoolsID="signalp" <https://bio.tools/api/t/?biotoolsID="signalp">`_

name                Search for tool name (quoted as needed)

                    `name=signalp <https://bio.tools/api/t/?name=signalp>`_
homepage            Exact search for tool homepage URL (**must** be quoted)

                    `homepage="http://cbs.dtu.dk/services/SignalP/" <https://bio.tools/api/t/?homepage="http://cbs.dtu.dk/services/SignalP/">`_
description         Search over tool description (quoted as needed)

                    `description="peptide cleavage" <https://bio.tools/api/t/?description="peptide%20cleavage">`_
version             Exact search for tool version (**must** be quoted)

                    `version="4.1" <https://bio.tools/api/t/?version="4.1">`_
topic               Search for EDAM Topic (term) (quoted as needed)

                    `topic="Proteomics" <https://bio.tools/api/t/?topic="Proteomics">`_

topicID             Exact search for EDAM Topic (URI): **must** be quoted

                    `topicID="topic_3510" <https://bio.tools/api/t/?topicID="topic_3510">`_
function            Fuzzy search over function (input, operation, output, note and command)

                    `function="Sequence analysis" <https://bio.tools/api/t/?function="Sequence%20analysis">`_
operation           Fuzzy search for EDAM Operation (term) (quoted as needed)

                    `operation="Sequence analysis" <https://bio.tools/api/t/?operation="Sequence%20analysis">`_
operationID         Exact search for EDAM Operation (ID) (**must** be quoted)

                    `operationID="operation_2403" <https://bio.tools/api/t/?operationID="operation_2403">`_
dataType            Fuzzy search over input and output for EDAM Data (term) (quoted as needed)

                    `dataType="Protein sequence" <https://bio.tools/api/t/?dataType="Protein%20sequence">`_
dataTypeID          Exact search over input and output for EDAM Data (ID) (**must** be quoted)

                    `dataTypeID="data_2976" <https://bio.tools/api/t/?dataTypeID="data_2976">`_
dataFormat          Fuzzy search over input and output for EDAM Format (term) (quoted as needed)

                    `dataFormat="FASTA" <https://bio.tools/api/t/?dataFormat="FASTA">`_
dataFormatID        Exact search over input and output for EDAM Format (ID) (**must** be quoted)

                    `dataFormatID="format_1929" <https://bio.tools/api/t/?dataFormatID="format_1929">`_
input               Fuzzy search over input for EDAM Data and Format (term) (quoted as needed)

                    `input="Protein sequence" <https://bio.tools/api/t/?input="Protein%20sequence">`_
inputID             Exact search over input for EDAM Data and Format (ID) (**must** be quoted)

                    `inputID="data_2976" <https://bio.tools/api/t/?inputID="data_2976">`_
inputDataType       Fuzzy search over input for EDAM Data (term) (quoted as needed)

                    `inputDataType="Protein sequence" <https://bio.tools/api/t/?inputDataType="Protein%20sequence">`_
inputDataTypeID     Exact search over input for EDAM Data (ID) (**must** be quoted)

                    `inputDataTypeID="data_2976" <https://bio.tools/api/t/?inputDataTypeID="data_2976">`_
inputDataFormat     Fuzzy search over input for EDAM Format (term) (quoted as needed)

                    `inputDataFormat="FASTA" <https://bio.tools/api/t/?inputDataFormat="FASTA">`_
inputDataFormatID   Exact search over input for EDAM Format (ID) (**must** be quoted)

                    `inputDataFormatID="format_1929" <https://bio.tools/api/t/?inputDataFormatID="format_1929">`_
output              Fuzzy search over output for EDAM Data and Format (term) (quoted as needed)

                    `output="Sequence alignment" <https://bio.tools/api/t/?output="Sequence%20alignment">`_
outputID            Exact search over output for EDAM Data and Format (ID) (**must** be quoted)

                    `outputID="data_0863" <https://bio.tools/api/t/?outputID="data_0863">`_
outputDataType      Fuzzy search over output for EDAM Data (term) (quoted as needed)

                    `outputDataType="Sequence alignment" <https://bio.tools/api/t/?outputDataType="Sequence%20alignment">`_
outputDataTypeID    Exact search over output for EDAM Data (ID) (**must** be quoted)

                    `outputDataTypeID="data_0863" <https://bio.tools/api/t/?outputDataTypeID="data_0863">`_
outputDataFormat    Fuzzy search over output for EDAM Format (term) (quoted as needed)

                    `outputDataFormat="ClustalW format" <https://bio.tools/api/t/?outputDataFormat="ClustalW%20format">`_
outputDataFormatID  Exact search over output for EDAM Format (ID) (**must** be quoted)

                    `outputDataFormatID="format_1982" <https://bio.tools/api/t/?outputDataFormatID="format_1982">`_
toolType            Exact search for tool type

                    `toolType="Command-line tool" <https://bio.tools/api/t/?toolType="Command-line%20tool">`_
collectionID        Exact search for tool collection (normally quoted)

                    `collectionID="Rare Disease" <https://bio.tools/api/t/?collectionID="Rare%20Disease">`_
maturity            Exact search for tool maturity

                    `maturity=Mature <https://bio.tools/api/t/?maturity=Mature>`_
operatingSystem     Exact search for tool operating system

                    `operatingSystem=Linux <https://bio.tools/api/t/?operatingSystem=Linux>`_
language            Exact search for programming language

                    `language=Java <https://bio.tools/api/t/?language=Java>`_
cost                Exact search for cost

                    `cost="Free of charge" <https://bio.tools/api/t/?cost="Free%20of%20charge">`_
license             Exact search for software or data usage license (quoted as needed)

                    `license="GPL-3.0" <https://bio.tools/api/t/?license="GPL-3.0">`_
accessibility       Exact search for tool accessibility

                    `accessibility="Open access" <https://bio.tools/api/t/?accessibility="Open%20access">`_
credit              Fuzzy search over credit (name, email, URL, ORCID iD, type of entity, type of role and note)

                    `credit="Henrik Nielsen" <https://bio.tools/api/t/?credit="Henrik%20Nielsen">`_
creditName          Exact search for name of credited entity

                    `creditName="Henrik Nielsen" <https://bio.tools/api/t/?creditName="Henrik%20Nielsen">`_
creditTypeRole      Exact search for role of credited entity

                    `creditTypeRole=Developer <https://bio.tools/api/t/?creditTypeRole=Developer>`_
creditTypeEntity    Exact search for type of credited entity

                    `creditTypeEntity="Funding agency" <https://bio.tools/api/t/?creditTypeEntity="Funding%20agency">`_
creditOrcidID       Exact search for ORCID iD of credited entity (**must** be quoted)

                    `creditOrcidID="0000-0001-5121-2036" <https://bio.tools/api/t/?creditOrcidID="0000-0001-5121-2036">`_
publication         Fuzzy search over publication (DOI, PMID, PMCID, publication type and tool version) (quoted as needed)

                    `publication=10.12688/f1000research.12974.1 <https://bio.tools/api/t/?publication=10.12688/f1000research.12974.1>`_
publicationID       Exact search for publication ID (DOI, PMID or PMCID) (**must** be quoted)

                    `publicationID="10.12688/f1000research.12974.1" <https://bio.tools/api/t/?publicationID="10.12688/f1000research.12974.1">`_
publicationType     Exact search for publication type

                    `publicationType=Primary <https://bio.tools/api/t/?publicationType=Primary>`_
publicationVersion  Exact search for tool version associated with a publication (**must** be quoted)

                    `publicationVersion="1.0" <https://bio.tools/api/t/?publicationVersion="1.0">`_
link                Fuzzy search over general link (URL, type and note) (quote as needed)

                    `link="Issue tracker" <https://bio.tools/api/t/?link="Issue%20tracker">`_
linkType            Exact search for type of information found at a link

                    `linkType="Issue tracker" <https://bio.tools/api/t/?linkType="Issue tracker">`_
documentation       Fuzzy search over documentation link (URL, type and note) (quote as needed)

                    `documentation=Manual <https://bio.tools/api/t/?documentation="User manual">`_
documentationType   Exact search for type of documentation

                    `documentationType=Manual <https://bio.tools/api/t/?documentationType="User manual">`_
download            Fuzzy search over download link (URL, type, version and note) (quote as needed)

                    `download=Binaries <https://bio.tools/api/t/?download=Binaries>`_
downloadType        Exact search for type of download

                    `downloadType=Binaries <https://bio.tools/api/t/?downloadType=Binaries>`_
downloadVersion     Exact search for tool version associated with a download (**must** be quoted)

                    `downloadVersion="1.0" <https://bio.tools/api/t/?downloadVersion="1.0">`_
otherID             Fuzzy search over alternate tool IDs (ID value, type of ID and version)

                    `otherID="rrid:SCR_015644" <https://bio.tools/api/t/?otherID="rrid:SCR_015644">`_

otherIDValue        Exact search for value of alternate tool ID (**must** be quoted)

                    `otherIDValue="rrid:SCR_015644" <https://bio.tools/api/t/?otherIDValue="rrid:SCR_015644">`_
otherIDType         Exact search for type of alternate tool ID

                    `otherIDType=RRID <https://bio.tools/api/t/?otherIDType=RRID>`_
otherIDVersion      Exact search for tool version associated with an alternate ID (**must** be quoted)

                    `otherIDVersion="1.0" <https://bio.tools/api/t/?otherIDVersion="1.0">`_
==================  ============================================================================================


The parameters are (currently) case-sensitive, e.g. you must use &biotoolsID= and not &biotoolsid

Values of the following parameters must be given in quotes to get sensible (or any) results:
homepage
version
topicID
operationID
dataTypeID
dataFormatID
inputID
inputDataTypeID
inputDataFormatID
outputID
outputDataTypeID
outputDataFormatID
creditOrcidID
publicationID
publicationVersion
downloadVersion
otherIDValue
otherIDVersion
e.g.
https://bio.tools/api/tool?topicID=”topic_3510”
Values of other parameters can be quoted or unquoted:
Unquoted values invoke a fuzzy word search: it will search for fuzzy matches of words in the search phrase, to the
target field
Quoted values invoke an exact phrase search; it will search for an exact match of the full-length of the search phrase,
to the target field (matches to target substrings are allowed)
e.g.
https://bio.tools/api/tool?biotoolsID=”blast” returns the tool with biotoolsID of “blast” (the “canonical” blast)
https://bio.tools/api/tool?biotoolsID=blast returns all tools with “blast” in their biotoolsID (all blast flavours)
"""


BIOTOOLS_SUMMARY_PROMPT = """
You have to answer this question in a clear and concise manner: {question} Be factual!\n\
You are a world leading bioinformatician who knows everything about bio.tools packages.\n\
Do not make up information, only use the provided information and mention how relevant the found information is based on
your knowledge about bio.tools.\n\
Here is the information relevant to the question found on the bio.tools web API:\n\
{context}
"""


class BioToolsQueryParameters(BaseModel):
    """Parameters for querying the bio.tools API."""

    base_url: str = Field(
        default="https://bio.tools/api/",
        description="Base URL for the BioTools API.",
    )
    endpoint: str = Field(
        ...,
        description="Specific API endpoint to hit. Example: 't/' for listing tools.",
    )
    biotoolsID: str | None = Field(  # noqa: N815
        None,
        description="Search for bio.tools tool ID (usually quoted - to get exact match)",
    )
    name: str | None = Field(
        None,
        description="Search for tool name (quoted as needed: quoted for exact match, unquoted for fuzzy search)",
    )
    homepage: str | None = Field(
        None,
        description="Exact search for tool homepage URL (**must** be quoted)",
    )
    description: str | None = Field(
        None,
        description="Search over tool description (quoted as needed)",
    )
    version: str | None = Field(
        None,
        description="Exact search for tool version (**must** be quoted)",
    )
    topic: str | None = Field(
        None,
        description="Search for EDAM Topic (term) (quoted as needed)",
    )
    topicID: str | None = Field(  # noqa: N815
        None,
        description="Exact search for EDAM Topic (URI): **must** be quoted",
    )
    function: str | None = Field(
        None,
        description="Fuzzy search over function (input, operation, output, note and command)",
    )
    operation: str | None = Field(
        None,
        description="Fuzzy search for EDAM Operation (term) (quoted as needed)",
    )
    operationID: str | None = Field(  # noqa: N815
        None,
        description="Exact search for EDAM Operation (ID) (**must** be quoted)",
    )
    dataType: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over input and output for EDAM Data (term) (quoted as needed)",
    )
    dataTypeID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over input and output for EDAM Data (ID) (**must** be quoted)",
    )
    dataFormat: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over input and output for EDAM Format (term) (quoted as needed)",
    )
    dataFormatID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over input and output for EDAM Format (ID) (**must** be quoted)",
    )
    input: str | None = Field(
        None,
        description="Fuzzy search over input for EDAM Data and Format (term) (quoted as needed)",
    )
    inputID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over input for EDAM Data and Format (ID) (**must** be quoted)",
    )
    inputDataType: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over input for EDAM Data (term) (quoted as needed)",
    )
    inputDataTypeID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over input for EDAM Data (ID) (**must** be quoted)",
    )
    inputDataFormat: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over input for EDAM Format (term) (quoted as needed)",
    )
    inputDataFormatID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over input for EDAM Format (ID) (**must** be quoted)",
    )
    output: str | None = Field(
        None,
        description="Fuzzy search over output for EDAM Data and Format (term) (quoted as needed)",
    )
    outputID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over output for EDAM Data and Format (ID) (**must** be quoted)",
    )
    outputDataType: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over output for EDAM Data (term) (quoted as needed)",
    )
    outputDataTypeID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over output for EDAM Data (ID) (**must** be quoted)",
    )
    outputDataFormat: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over output for EDAM Format (term) (quoted as needed)",
    )
    outputDataFormatID: str | None = Field(  # noqa: N815
        None,
        description="Exact search over output for EDAM Format (ID) (**must** be quoted)",
    )
    toolType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool type",
    )
    collectionID: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool collection (normally quoted)",
    )
    maturity: str | None = Field(
        None,
        description="Exact search for tool maturity",
    )
    operatingSystem: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool operating system",
    )
    language: str | None = Field(
        None,
        description="Exact search for programming language",
    )
    cost: str | None = Field(
        None,
        description="Exact search for cost",
    )
    license: str | None = Field(
        None,
        description="Exact search for software or data usage license (quoted as needed)",
    )
    accessibility: str | None = Field(
        None,
        description="Exact search for tool accessibility",
    )
    credit: str | None = Field(
        None,
        description="Fuzzy search over credit (name, email, URL, ORCID iD, type of entity, type of role and note)",
    )
    creditName: str | None = Field(  # noqa: N815
        None,
        description="Exact search for name of credited entity",
    )
    creditTypeRole: str | None = Field(  # noqa: N815
        None,
        description="Exact search for role of credited entity",
    )
    creditTypeEntity: str | None = Field(  # noqa: N815
        None,
        description="Exact search for type of credited entity",
    )
    creditOrcidID: str | None = Field(  # noqa: N815
        None,
        description="Exact search for ORCID iD of credited entity (**must** be quoted)",
    )
    publication: str | None = Field(
        None,
        description=(
            "Fuzzy search over publication (DOI, PMID, PMCID, publication type and tool version) (quoted as needed)"
        ),
    )
    publicationID: str | None = Field(  # noqa: N815
        None,
        description="Exact search for publication ID (DOI, PMID or PMCID) (**must** be quoted)",
    )
    publicationType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for publication type",
    )
    publicationVersion: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool version associated with a publication (**must** be quoted)",
    )
    link: str | None = Field(
        None,
        description="Fuzzy search over general link (URL, type and note) (quote as needed)",
    )
    linkType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for type of information found at a link",
    )
    documentation: str | None = Field(
        None,
        description="Fuzzy search over documentation link (URL, type and note) (quote as needed)",
    )
    documentationType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for type of documentation",
    )
    download: str | None = Field(
        None,
        description="Fuzzy search over download link (URL, type, version and note) (quote as needed)",
    )
    downloadType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for type of download",
    )
    downloadVersion: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool version associated with a download (**must** be quoted)",
    )
    otherID: str | None = Field(  # noqa: N815
        None,
        description="Fuzzy search over alternate tool IDs (ID value, type of ID and version)",
    )
    otherIDValue: str | None = Field(  # noqa: N815
        None,
        description="Exact search for value of alternate tool ID (**must** be quoted)",
    )
    otherIDType: str | None = Field(  # noqa: N815
        None,
        description="Exact search for type of alternate tool ID",
    )
    otherIDVersion: str | None = Field(  # noqa: N815
        None,
        description="Exact search for tool version associated with an alternate ID (**must** be quoted)",
    )
    question_uuid: str | None = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="Unique identifier for the question.",
    )


class BioToolsQueryBuilder(BaseQueryBuilder):
    """A class for building an BioToolsQuery object."""

    def create_runnable(
        self,
        query_parameters: "BioToolsQueryParameters",
        conversation: "Conversation",
    ) -> Callable:
        """Create a runnable object for executing queries.

        Create runnable using the LangChain `create_structured_output_runnable`
        method.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """
        return create_structured_output_runnable(
            output_schema=query_parameters,
            llm=conversation.chat,
            prompt=self.structured_output_prompt,
        )

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list[BioToolsQueryParameters]:
        """Generate an BioToolsQuery object.

        Generate a BioToolsQuery object based on the given question, prompt,
        and BioChatter conversation. Uses a Pydantic model to define the API
        fields.  Creates a runnable that can be invoked on LLMs that are
        qualified to parameterise functions.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                BioToolsQuery.

        Returns:
        -------
            BioToolsQueryParameters: the parameterised query object (Pydantic
                model)

        """
        runnable = self.create_runnable(
            query_parameters=BioToolsQueryParameters,
            conversation=conversation,
        )
        oncokb_call_obj = runnable.invoke(
            {
                "input": f"Answer:\n{question} based on:\n {BIOTOOLS_QUERY_PROMPT}",
            },
        )
        oncokb_call_obj.question_uuid = str(uuid.uuid4())
        return [oncokb_call_obj]


class BioToolsFetcher(BaseFetcher):
    """A class for retrieving API results from BioTools.

    Retrieves API results given a parameterized BioToolsQuery.
    """

    def __init__(self, api_token: str = "demo") -> None:  # noqa: S107
        """Initialise the BioToolsFetcher.

        Args:
        ----
            api_token: The API token for the BioTools API.

        """
        self.headers = {
            "Authorization": f"Bearer {api_token}",
            "Accept": "application/json",
        }
        self.base_url = "https://bio.tools/api"

    def fetch_results(
        self,
        request_data: list[BioToolsQueryParameters],
        retries: int | None = 3,  # noqa: ARG002
    ) -> str:
        """Submit the BioTools query and fetch the results directly.

        No multi-step procedure, thus no wrapping of submission and retrieval in
        this case.

        Args:
        ----
            request_data: List of BioToolsQuery objects (Pydantic models)
                containing the BioTools query parameters.

            retries: The number of retries to fetch the results.

        Returns:
        -------
            str: The results of the BioTools query.

        """
        # For now, we only use the first query in the list
        query = request_data[0]

        # Submit the query and get the URL
        params = query.dict(exclude_unset=True)
        endpoint = params.pop("endpoint")
        params.pop("question_uuid")
        full_url = f"{self.base_url}/{endpoint}"
        response = requests.get(full_url, headers=self.headers, params=params, timeout=30)
        response.raise_for_status()

        # Fetch the results from the URL
        results_response = requests.get(response.url, headers=self.headers, timeout=30)
        results_response.raise_for_status()

        return results_response.text


class BioToolsInterpreter(BaseInterpreter):
    """A class for interpreting BioTools results."""

    def summarise_results(
        self,
        question: str,
        conversation_factory: Callable,
        response_text: str,
    ) -> str:
        """Extract the answer from the BLAST results.

        Args:
        ----
            question (str): The question to be answered.
            conversation_factory: A BioChatter conversation object.
            response_text (str): The response.text returned by bio.tools.

        Returns:
        -------
            str: The extracted answer from the BLAST results.

        """
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a world class bioinformatician who knows "
                    "everything about bio.tools packages and the "
                    "bioinformatics ecosystem. Your task is to interpret "
                    "results from BioTools API calls and summarise "
                    "them for the user.",
                ),
                ("user", "{input}"),
            ],
        )
        summary_prompt = BIOTOOLS_SUMMARY_PROMPT.format(
            question=question,
            context=response_text,
        )
        output_parser = StrOutputParser()
        conversation = conversation_factory()
        chain = prompt | conversation.chat | output_parser
        return chain.invoke({"input": {summary_prompt}})



from collections.abc import Callable
from typing import Literal

from langchain_core.messages import HumanMessage, SystemMessage
from langchain_google_genai import ChatGoogleGenerativeAI

from biochatter.llm_connect.available_models import TOOL_CALLING_MODELS
from biochatter.llm_connect.conversation import Conversation


class GeminiConversation(Conversation):
    """Conversation class for the Google Gemini model."""

    def __init__(
        self,
        model_name: str,
        prompts: dict,
        correct: bool = False,
        split_correction: bool = False,
        tools: list[Callable] = None,
        tool_call_mode: Literal["auto", "text"] = "auto",
    ) -> None:
        """Initialise the GeminiConversation class.

        Connect to Google's Gemini API and set up a conversation with the user.
        Also initialise a second conversational agent to provide corrections to
        the model output, if necessary.

        Args:
        ----
            model_name (str): The name of the model to use.

            prompts (dict): A dictionary of prompts to use for the conversation.

            correct (bool): Whether to correct the model output.

            split_correction (bool): Whether to correct the model output by
                splitting the output into sentences and correcting each
                sentence individually.

            tools (list[Callable]): List of tool functions to use with the model.

            tool_call_mode (str): The mode to use for tool calls.
                "auto": Automatically call tools.
                "text": Only return text output of the tool call.

        """
        super().__init__(
            model_name=model_name,
            prompts=prompts,
            correct=correct,
            split_correction=split_correction,
            tools=tools,
            tool_call_mode=tool_call_mode,
        )

        self.ca_model_name = "gemini-2.0-flash"

    def set_api_key(self, api_key: str, user: str | None = None) -> bool:
        """Set the API key for the Google Gemini API.

        If the key is valid, initialise the conversational agent. Optionally set
        the user for usage statistics.

        Args:
        ----
            api_key (str): The API key for the Google Gemini API.

            user (str, optional): The user for usage statistics. If provided and
                equals "community", will track usage stats.

        Returns:
        -------
            bool: True if the API key is valid, False otherwise.

        """
        self.user = user

        try:
            self.chat = ChatGoogleGenerativeAI(
                model=self.model_name,
                temperature=0,
                google_api_key=api_key,
            )
            self.ca_chat = ChatGoogleGenerativeAI(
                model=self.ca_model_name,
                temperature=0,
                google_api_key=api_key,
            )

            # if binding happens here, tools will be available for all messages
            if self.tools:
                self.bind_tools(self.tools)

            return True

        except Exception:  # Google Genai doesn't expose specific exception types
            self._chat = None
            self._ca_chat = None
            return False

    def _primary_query(self, tools: list[Callable] | None = None, **kwargs) -> tuple:
        """Query the Google Gemini API with the user's message.

        Return the response using the message history (flattery system messages,
        prior conversation) as context. Correct the response if necessary.

        Args:
        ----
            tools (list[Callable]): The tools to use for the query. Tools
            passed at this step are used only for this message and not stored
            as part of the conversation object.

            **kwargs: Additional keyword arguments.

        Returns:
        -------
            tuple: A tuple containing the response from the Gemini API and
                the token usage.

        """
        # bind tools to the chat if provided in the query
        chat = self.chat.bind_tools(tools) if (tools and self.model_name in TOOL_CALLING_MODELS) else self.chat

        try:
            response = chat.invoke(self.messages)
        except Exception as e:
            return str(e), None

        # Process tool calls if present
        if response.tool_calls:
            msg = self._process_tool_calls(response.tool_calls, tools, response.content)
        else:
            msg = response.content
            self.append_ai_message(msg)

        token_usage = response.usage_metadata["total_tokens"]

        return msg, token_usage

    def _correct_response(self, msg: str) -> str:
        """Correct the response from the Gemini API.

        Send the response to a secondary language model. Optionally split the
        response into single sentences and correct each sentence individually.
        Update usage stats.

        Args:
        ----
            msg (str): The response from the Gemini API.

        Returns:
        -------
            str: The corrected response (or OK if no correction necessary).

        """
        ca_messages = self.ca_messages.copy()

        ca_messages.append(
            SystemMessage(
                content="If there is nothing to correct, please respond with just 'OK', and nothing else!",
            ),
        )

        ca_messages.append(
            HumanMessage(
                content=msg,
            ),
        )

        response = self.ca_chat.invoke(ca_messages)

        correction = response.content
        token_usage = response.usage_metadata["total_tokens"]

        return correction



"""Abstract base classes for API interaction components.

Provides base classes for query builders, fetchers, and interpreters used in
API interactions and result processing.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable

from langchain_core.prompts import ChatPromptTemplate
from pydantic import BaseModel, ConfigDict, Field, create_model

from biochatter.llm_connect import Conversation


class BaseQueryBuilder(ABC):
    """An abstract base class for query builders."""

    @property
    def structured_output_prompt(self) -> ChatPromptTemplate:
        """Define a structured output prompt template.

        This provides a default implementation for an API agent that can be
        overridden by subclasses to return a ChatPromptTemplate-compatible
        object.
        """
        return ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a world class algorithm for extracting information in structured formats.",
                ),
                (
                    "human",
                    "Use the given format to extract information from the following input: {input}",
                ),
                ("human", "Tip: Make sure to answer in the correct format"),
            ],
        )

    @abstractmethod
    def create_runnable(
        self,
        query_parameters: "BaseModel",
        conversation: "Conversation",
    ) -> Callable:
        """Create a runnable object for executing queries.

        Must be implemented by subclasses. Should use the LangChain
        `create_structured_output_runnable` method to generate the Callable.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """

    @abstractmethod
    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list[BaseModel]:
        """Parameterise a query object.

        Parameterises a Pydantic model with the fields of the API based on the
        given question using a BioChatter conversation instance. Must be
        implemented by subclasses.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The BioChatter conversation object containing the LLM
                that should parameterise the query.

        Returns:
        -------
            A list containing one or more parameterised instance(s) of the query
            object (Pydantic BaseModel).

        """


class BaseFetcher(ABC):
    """Abstract base class for fetchers.

    A fetcher is responsible for submitting queries (in systems where
    submission and fetching are separate) and fetching and saving results of
    queries. It has to implement a `fetch_results()` method, which can wrap a
    multi-step procedure to submit and retrieve. Should implement retry method to
    account for connectivity issues or processing times.
    """

    @abstractmethod
    def fetch_results(
        self,
        query_models: list[BaseModel],
        retries: int | None = 3,
    ):
        """Fetch results by submitting a query.

        Can implement a multi-step procedure if submitting and fetching are
        distinct processes (e.g., in the case of long processing times as in the
        case of BLAST).

        Args:
        ----
            query_models: list of Pydantic models describing the parameterised
                queries

        """


class BaseInterpreter(ABC):
    """Abstract base class for result interpreters.

    The interpreter is aware of the nature and structure of the results and can
    extract and summarise information from them.
    """

    @abstractmethod
    def summarise_results(
        self,
        question: str,
        conversation_factory: Callable,
        response_text: str,
    ) -> str:
        """Summarise an answer based on the given parameters.

        Args:
        ----
            question (str): The question that was asked.

            conversation_factory (Callable): A function that creates a
                BioChatter conversation.

            response_text (str): The response.text returned from the request.

        Returns:
        -------
            A summary of the answer.

        Todo:
        ----
            Genericise (remove file path and n_lines parameters, and use a
            generic way to get the results). The child classes should manage the
            specifics of the results.

        """


class BaseAPIModel(BaseModel):
    """A base class for all API models.

    Includes default fields `uuid` and `method_name`.
    """

    uuid: str | None = Field(
        None,
        description="Unique identifier for the model instance",
    )
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseTools:
    """Abstract base class for tools."""

    def make_pydantic_tools(self) -> list[BaseAPIModel]:
        """Uses pydantics create_model to create a list of pydantic tools from a dictionary of parameters"""
        tools = []
        for func_name, tool_params in self.tools_params.items():
            tools.append(create_model(func_name, **tool_params, __base__=BaseAPIModel))
        return tools



"""Scanpy Preprocessing (scanpy.pp) Query Builder.

TODO: not sure if the approach below is functional yet.
"""

from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Optional, Union

from langchain_core.output_parsers import PydanticToolsParser
from pydantic import BaseModel, Field

from biochatter.api_agent.base.agent_abc import (
    BaseAPIModel,
    BaseQueryBuilder,
    BaseTools,
)
from biochatter.llm_connect import Conversation

if TYPE_CHECKING:
    from biochatter.llm_connect import Conversation

SCANPY_PL_QUERY_PROMPT = """
Scanpy Preprocessing (scanpy.pp) Query Guide

You are a world-class algorithm for creating queries in structured formats. Your task is to use the Python API of scanpy
to answer questions about the scanpy.pp (preprocessing) module. All function calls should be prefixed with `scanpy.pp.`.
For example, to normalize the data, you should use scanpy.pp.normalize_total.

Use the following documentation to craft queries for preprocessing tasks:
Preprocessing Functions in scanpy.pp

These are the primary preprocessing functions in scanpy.pp used for data cleaning, transformation, and filtering.

    pp.filter_cells
    Filters cells based on the minimum and maximum number of counts or genes.

    pp.filter_genes
    Filters genes based on the minimum and maximum number of counts or cells in which the gene is expressed.

    pp.normalize_total
    Normalizes the data by scaling each cell to a specified total count.

    pp.log1p
    Applies a natural logarithm transformation to the data (adds 1 before log transformation).

    pp.regress_out
    Removes the effects of unwanted sources of variation by regressing out specific variables.

    pp.scale
    Scales the data by zero-centering and (optionally) scaling each feature.

    pp.subsample
    Subsamples the data by randomly selecting a fraction of the observations or by a fixed number of observations.

    pp.highly_variable_genes
    Identifies and selects highly variable genes based on mean and dispersion criteria.

    pp.calculate_qc_metrics
    Computes quality control metrics such as the number of genes detected per cell, total counts per cell, and more.

Specialized Preprocessing Methods

These functions are used for specialized preprocessing workflows:

    pp.recipe_zhen17
    Implements a preprocessing recipe as described in Zheng et al., 2017 for single-cell RNA-seq data.

    pp.recipe_weinreb17
    Implements a preprocessing recipe as described in Weinreb et al., 2017 for single-cell RNA-seq data.

    pp.recipe_seurat
    Implements a preprocessing recipe for integration with the Seurat workflow.

    pp.combat
    Removes batch effects using the ComBat method.

    pp.scrublet
    Simulates and detects doublets in the dataset.

General Functions

    pp.dummy
    Placeholder function for dummy operations or custom preprocessing routines.

    pp.projection
    Projects the data into a new space after dimensionality reduction techniques like PCA.

Use the provided documentation to craft precise queries for any preprocessing needs in Scanpy. Ensure that your function
call starts with scanpy.pp. and that you include relevant parameters based on the query.

This prompt guides the user to query the scanpy.pp module for preprocessing tasks, assisting with the construction of
specific preprocessing operations, such as filtering, normalization, scaling, and more.
"""


class ScanpyPpFuncs(BaseTools):
    """Scanpy Preprocessing (scanpy.pp) Query Builder."""

    tools_params = {}

    tools_params["filter_cells"] = {
        "data": (str, Field(..., description="The (annotated) data matrix.")),
        "min_counts": (Optional[int], Field(None, description="Minimum counts per cell.")),
        "min_genes": (Optional[int], Field(None, description="Minimum genes expressed in a cell.")),
        "max_counts": (Optional[int], Field(None, description="Maximum counts per cell.")),
        "max_genes": (Optional[int], Field(None, description="Maximum genes expressed in a cell.")),
        "inplace": (bool, Field(True, description="Whether to modify the data in place.")),
    }

    tools_params["filter_genes"] = {
        "data": (str, Field(..., description="The (annotated) data matrix.")),
        "min_counts": (Optional[int], Field(None, description="Minimum counts per gene.")),
        "min_cells": (Optional[int], Field(None, description="Minimum number of cells expressing the gene.")),
        "max_counts": (Optional[int], Field(None, description="Maximum counts per gene.")),
        "max_cells": (Optional[int], Field(None, description="Maximum number of cells expressing the gene.")),
        "inplace": (bool, Field(True, description="Whether to modify the data in place.")),
    }

    tools_params["highly_variable_genes"] = {
        "adata": (str, Field(..., description="Annotated data matrix.")),
        "n_top_genes": (Optional[int], Field(None, description="Number of highly-variable genes to keep.")),
        "min_mean": (float, Field(0.0125, description="Minimum mean expression for highly-variable genes.")),
        "max_mean": (float, Field(3, description="Maximum mean expression for highly-variable genes.")),
        "flavor": (str, Field("seurat", description="Method for identifying highly-variable genes.")),
        "inplace": (bool, Field(True, description="Whether to place metrics in .var or return them.")),
    }

    tools_params["log1p"] = {
        "data": (str, Field(..., description="The data matrix.")),
        "base": (Optional[float], Field(None, description="Base of the logarithm.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
        "chunked": (Optional[bool], Field(None, description="Process data in chunks.")),
    }

    tools_params["pca"] = {
        "data": (str, Field(..., description="The (annotated) data matrix.")),
        "n_comps": (Optional[int], Field(None, description="Number of principal components to compute.")),
        "layer": (Optional[str], Field(None, description="Element of layers to use for PCA.")),
        "zero_center": (bool, Field(True, description="Whether to zero-center the data.")),
        "svd_solver": (Optional[str], Field(None, description="SVD solver to use.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
    }

    tools_params["normalize_total"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "target_sum": (Optional[float], Field(None, description="Target sum after normalization.")),
        "exclude_highly_expressed": (bool, Field(False, description="Whether to exclude highly expressed genes.")),
        "inplace": (bool, Field(True, description="Whether to update adata or return normalized data.")),
    }

    tools_params["regress_out"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "keys": (Union[str, Collection[str]], Field(..., description="Keys for regression.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
    }

    tools_params["scale"] = {
        "data": (str, Field(..., description="The data matrix.")),
        "zero_center": (bool, Field(True, description="Whether to zero-center the data.")),
        "copy": (bool, Field(False, description="Whether to perform operation inplace.")),
    }

    tools_params["subsample"] = {
        "data": (str, Field(..., description="The data matrix.")),
        "fraction": (Optional[float], Field(None, description="Fraction of observations to subsample.")),
        "n_obs": (Optional[int], Field(None, description="Number of observations to subsample.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
    }

    tools_params["downsample_counts"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "counts_per_cell": (Optional[int | str], Field(None, description="Target total counts per cell.")),
        "replace": (bool, Field(False, description="Whether to sample with replacement.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
    }

    tools_params["combat"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "key": (str, Field("batch", description="Key for batch effect removal.")),
        "inplace": (bool, Field(True, description="Whether to replace the data inplace.")),
    }

    tools_params["scrublet"] = {
        "adata": (str, Field(..., description="Annotated data matrix.")),
        "sim_doublet_ratio": (float, Field(2.0, description="Number of doublets to simulate.")),
        "threshold": (Optional[float], Field(None, description="Doublet score threshold.")),
        "copy": (bool, Field(False, description="If True, return a copy_param of the data.")),
    }

    tools_params["scrublet_simulate_doublets"] = {
        "adata": (str, Field(..., description="Annotated data matrix.")),
        "sim_doublet_ratio": (float, Field(2.0, description="Number of doublets to simulate.")),
        "random_seed": (int, Field(0, description="Random seed for reproducibility.")),
    }
    tools_params["calculate_qc_metrics"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "expr_type": (str, Field("counts", description="Name of kind of values in X.")),
        "var_type": (str, Field("genes", description="The kind of thing the variables are.")),
        "qc_vars": (
            Collection[str],
            Field(
                (),
                description="Keys for boolean columns of .var which identify variables you could want to control for (e.g., “ERCC” or “mito”).",
            ),
        ),
        "percent_top": (
            Collection[int],
            Field(
                (50, 100, 200, 500),
                description="List of ranks at which cumulative proportion of expression will be reported as a percentage.",
            ),
        ),
        "layer": (
            Optional[str],
            Field(None, description="If provided, use adata.layers[layer] for expression values instead of adata.X."),
        ),
        "use_raw": (
            bool,
            Field(False, description="If True, use adata.raw.X for expression values instead of adata.X."),
        ),
        "inplace": (bool, Field(False, description="Whether to place calculated metrics in adata’s .obs and .var.")),
        "log1p": (bool, Field(True, description="Set to False to skip computing log1p transformed annotations.")),
    }

    tools_params["recipe_zheng17"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "n_top_genes": (int, Field(1000, description="Number of genes to keep.")),
        "log": (bool, Field(True, description="Take logarithm of the data.")),
        "plot": (bool, Field(False, description="Show a plot of the gene dispersion vs. mean relation.")),
        "copy": (bool, Field(False, description="Return a copy of adata instead of updating it.")),
    }

    tools_params["recipe_weinreb17"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "log": (bool, Field(True, description="Logarithmize the data?")),
        "mean_threshold": (float, Field(0.01, description="Mean expression threshold for gene selection.")),
        "cv_threshold": (float, Field(2, description="Coefficient of variation threshold for gene selection.")),
        "n_pcs": (int, Field(50, description="Number of principal components to compute.")),
        "svd_solver": (str, Field("randomized", description="SVD solver to use for PCA.")),
        "random_state": (int, Field(0, description="Random seed for reproducibility.")),
        "copy": (
            bool,
            Field(False, description="Return a copy if true, otherwise modifies the original adata object."),
        ),
    }

    tools_params["recipe_seurat"] = {
        "adata": (str, Field(..., description="The annotated data matrix.")),
        "log": (bool, Field(True, description="Logarithmize the data?")),
        "plot": (bool, Field(False, description="Show a plot of the gene dispersion vs. mean relation.")),
        "copy": (
            bool,
            Field(False, description="Return a copy if true, otherwise modifies the original adata object."),
        ),
    }

    def __init__(self, tools_params: dict = tools_params) -> None:
        """Initialise the ScanpyPpFuncs class."""
        super().__init__()
        self.tools_params = tools_params


class ScanpyPpQueryBuilder(BaseQueryBuilder):
    """A class for building a ScanpyPp query object."""

    def create_runnable(
        self,
        query_parameters: list["BaseAPIModel"],
        conversation: "Conversation",
    ) -> Callable:
        """Create a runnable object for executing queries.

        Create runnable using the LangChain `create_structured_output_runnable`
        method.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """
        runnable = conversation.chat.bind_tools(query_parameters)
        return runnable | PydanticToolsParser(tools=query_parameters)

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list["BaseModel"]:
        """Generate a ScanpyPp query object.

        Generates the object based on the given question, prompt, and
        BioChatter conversation. Uses a Pydantic model to define the API fields.
        Creates a runnable that can be invoked on LLMs that are qualified to
        parameterise functions.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                ScanpyPpQuery.

        Returns:
        -------
            ScanpyPpQuery: the parameterised query object (Pydantic model)

        """
        tool_maker = ScanpyPpFuncs()
        tools = tool_maker.make_pydantic_tools()
        runnable = self.create_runnable(conversation=conversation, query_parameters=tools)
        return runnable.invoke(
            question,
        )



"""Module for interacting with the `scanpy` API for plotting (`pl`)."""

import uuid
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from langchain_core.output_parsers import PydanticToolsParser
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from biochatter.llm_connect import Conversation

from biochatter.api_agent.base.agent_abc import BaseAPIModel, BaseQueryBuilder

SCANPY_PL_QUERY_PROMPT = """
You are a world class algorithm for creating queries in structured formats.
Your task is to use the Python API of `scanpy` to answer questions about the `scanpy` plotting module.
You should prefix the function calls with `scanpy.pl.`, for instance, for a scatter plot, you should use
`scanpy.pl.scatter`.

Extract information from the following documentation to answer the user's question:

The plotting module scanpy.pl largely parallels the tl.* and a few of the pp.* functions. For most tools and for some
preprocessing functions, you will find a plotting function with the same name.

Generic
pl.scatter
Scatter plot along observations or variables axes.
pl.heatmap
Heatmap of the expression values of genes.
pl.dotplot
Makes a dot plot of the expression values of var_names.
pl.tracksplot
In this type of plot each var_name is plotted as a filled line plot where the y values correspond to the var_name values
and x is each of the cells.
pl.violin
Violin plot.
pl.stacked_violin
Stacked violin plots.
pl.matrixplot
Creates a heatmap of the mean expression values per group of each var_names.
pl.clustermap
Hierarchically-clustered heatmap.
pl.ranking
Plot rankings.
pl.dendrogram
Plots a dendrogram of the categories defined in groupby.
Classes
These classes allow fine tuning of visual parameters.

pl.DotPlot
Allows the visualization of two values that are encoded as dot size and color.
pl.MatrixPlot
Allows the visualization of values using a color map.
pl.StackedViolin
Stacked violin plots.
Preprocessing
Methods for visualizing quality control and results of preprocessing functions.

pl.highest_expr_genes
Fraction of counts assigned to each gene over all cells.
pl.filter_genes_dispersion
Plot dispersions versus means for genes.
pl.highly_variable_genes
Plot dispersions or normalized variance versus means for genes.
pl.scrublet_score_distribution
Plot histogram of doublet scores for observed transcriptomes and simulated doublets.
Tools
Methods that extract and visualize tool-specific annotation in an AnnData object. For any method in module tl, there is
a method with the same name in pl.

PCA

pl.pca
Scatter plot in PCA coordinates.
pl.pca_loadings
Rank genes according to contributions to PCs.
pl.pca_variance_ratio
Plot the variance ratio.
pl.pca_overview
Plot PCA results.
Embeddings

pl.tsne
Scatter plot in tSNE basis.
pl.umap
Scatter plot in UMAP basis.
pl.diffmap
Scatter plot in Diffusion Map basis.
pl.draw_graph
Scatter plot in graph-drawing basis.
pl.spatial
Scatter plot in spatial coordinates.
pl.embedding
Scatter plot for user specified embedding basis (e.g. umap, pca, etc).
Compute densities on embeddings.

pl.embedding_density
Plot the density of cells in an embedding (per condition).
Branching trajectories and pseudotime, clustering

Visualize clusters using one of the embedding methods passing color='louvain'.

pl.dpt_groups_pseudotime
Plot groups and pseudotime.
pl.dpt_timeseries
Heatmap of pseudotime series.
pl.paga
Plot the PAGA graph through thresholding low-connectivity edges.
pl.paga_path
Gene expression and annotation changes along paths in the abstracted graph.
pl.paga_compare
Scatter and PAGA graph side-by-side.
Marker genes

pl.rank_genes_groups
Plot ranking of genes.
pl.rank_genes_groups_violin
Plot ranking of genes for all tested comparisons.
pl.rank_genes_groups_stacked_violin
Plot ranking of genes using stacked_violin plot (see stacked_violin())
pl.rank_genes_groups_heatmap
Plot ranking of genes using heatmap plot (see heatmap())
pl.rank_genes_groups_dotplot
Plot ranking of genes using dotplot plot (see dotplot())
pl.rank_genes_groups_matrixplot
Plot ranking of genes using matrixplot plot (see matrixplot())
pl.rank_genes_groups_tracksplot
Plot ranking of genes using heatmap plot (see heatmap())
Simulations

pl.sim
Plot results of simulation.
"""


class ScanpyPlScatterQueryParameters(BaseModel):
    """Parameters for querying the scanpy `pl.scatter` API."""

    method_name: str = Field(
        default="sc.pl.scatter",
        description="The name of the method to call.",
    )
    question_uuid: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="Unique identifier for the question.",
    )
    adata: str = Field(description="Annotated data matrix.")
    x: str | None = Field(default=None, description="x coordinate.")
    y: str | None = Field(default=None, description="y coordinate.")
    color: str | tuple[float, ...] | list[str | tuple[float, ...]] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes, or a hex color specification.",
    )
    use_raw: bool | None = Field(
        default=None,
        description="Whether to use raw attribute of adata. Defaults to True if .raw is present.",
    )
    layers: str | list[str] | None = Field(
        default=None,
        description="Layer(s) to use from adata's layers attribute.",
    )
    basis: str | None = Field(
        default=None,
        description="String that denotes a plotting tool that computed coordinates (e.g., 'pca', 'tsne', 'umap').",
    )
    sort_order: bool = Field(
        default=True,
        description="For continuous annotations used as color parameter, plot data points with higher values on top.",
    )
    groups: str | list[str] | None = Field(
        default=None,
        description="Restrict to specific categories in categorical observation annotation.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot ('2d' or '3d').",
    )
    legend_loc: str | None = Field(
        default="right margin",
        description="Location of legend ('none', 'right margin', 'on data', etc.).",
    )
    size: int | float | None = Field(
        default=None,
        description="Point size. If None, automatically computed as 120000 / n_cells.",
    )
    color_map: str | None = Field(
        default=None,
        description="Color map to use for continuous variables (e.g., 'magma', 'viridis').",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot, do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If True or a str, save the figure. String is appended to default filename.",
    )


### Embeddings
class ScanpyPlPcaQueryParameters(BaseModel):
    """Parameters for querying the scanpy `pl.pca` API."""

    method_name: str = Field(
        default="sc.pl.pca",
        description="The name of the method to call.",
    )
    question_uuid: str | None = Field(
        default=None,
        description="Unique identifier for the question.",
    )
    adata: str = Field(
        ...,
        description="Annotated data matrix.",
    )
    color: str | list[str] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )
    components: str | list[str] = Field(
        default="1,2",
        description="For example, ['1,2', '2,3']. To plot all available components use 'all'.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot.",
    )
    legend_loc: str = Field(
        default="right margin",
        description="Location of legend.",
    )
    legend_fontsize: int | float | str | None = Field(
        default=None,
        description="Font size for legend.",
    )
    legend_fontweight: int | str | None = Field(
        default=None,
        description="Font weight for legend.",
    )
    color_map: str | None = Field(
        default=None,
        description="String denoting matplotlib color map.",
    )
    palette: str | list[str] | dict | None = Field(
        default=None,
        description="Colors to use for plotting categorical annotation groups.",
    )
    frameon: bool | None = Field(
        default=None,
        description="Draw a frame around the scatter plot.",
    )
    size: int | float | None = Field(
        default=None,
        description="Point size. If `None`, is automatically computed as 120000 / n_cells.",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot, do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If `True` or a `str`, save the figure.",
    )
    ax: str | None = Field(
        default=None,
        description="A matplotlib axes object.",
    )
    return_fig: bool = Field(
        default=False,
        description="Return the matplotlib figure object.",
    )
    marker: str | None = Field(
        default=".",
        description="Marker symbol.",
    )
    annotate_var_explained: bool = Field(
        default=False,
        description="Annotate the percentage of explained variance.",
    )


class ScanpyPlTsneQueryParameters(BaseModel):
    """Parameters for querying the Scanpy `pl.tsne` API."""

    method_name: str = Field(
        default="sc.pl.tsne",
        description="The name of the method to call.",
    )
    question_uuid: str | None = Field(
        default=None,
        description="Unique identifier for the question.",
    )
    adata: str = Field(
        ...,
        description="Annotated data matrix.",
    )
    color: str | list[str] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )
    gene_symbols: str | None = Field(
        default=None,
        description="Column name in `.var` DataFrame that stores gene symbols.",
    )
    use_raw: bool | None = Field(
        default=None,
        description="Use `.raw` attribute of `adata` for coloring with gene expression.",
    )
    sort_order: bool = Field(
        default=True,
        description="Plot data points with higher values on top for continuous annotations.",
    )
    edges: bool = Field(
        default=False,
        description="Show edges.",
    )
    edges_width: float = Field(
        default=0.1,
        description="Width of edges.",
    )
    edges_color: str | list[float] | list[str] = Field(
        default="grey",
        description="Color of edges.",
    )
    neighbors_key: str | None = Field(
        default=None,
        description="Key for neighbors connectivities.",
    )
    arrows: bool = Field(
        default=False,
        description="Show arrows (deprecated in favor of `scvelo.pl.velocity_embedding`).",
    )
    arrows_kwds: dict[str, Any] | None = Field(
        default=None,
        description="Arguments passed to `quiver()`.",
    )
    groups: str | None = Field(
        default=None,
        description="Restrict to specific categories in categorical observation annotation.",
    )
    components: str | list[str] | None = Field(
        default=None,
        description="Components to plot, e.g., ['1,2', '2,3']. Use 'all' to plot all available components.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot ('2d' or '3d').",
    )
    legend_loc: str = Field(
        default="right margin",
        description="Location of legend.",
    )
    legend_fontsize: int | float | str | None = Field(
        default=None,
        description="Font size for legend.",
    )
    legend_fontweight: int | str = Field(
        default="bold",
        description="Font weight for legend.",
    )
    legend_fontoutline: int | None = Field(
        default=None,
        description="Line width of the legend font outline in pt.",
    )
    size: float | list[float] | None = Field(
        default=None,
        description="Point size. If `None`, computed as 120000 / n_cells.",
    )
    color_map: str | Any | None = Field(
        default=None,
        description="Color map for continuous variables.",
    )
    palette: str | list[str] | Any | None = Field(
        default=None,
        description="Colors for plotting categorical annotation groups.",
    )
    na_color: str | tuple[float, ...] = Field(
        default="lightgray",
        description="Color for null or masked values.",
    )
    na_in_legend: bool = Field(
        default=True,
        description="Include missing values in the legend.",
    )
    frameon: bool | None = Field(
        default=None,
        description="Draw a frame around the scatter plot.",
    )
    vmin: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Lower limit of the color scale.",
    )
    vmax: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Upper limit of the color scale.",
    )
    vcenter: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Center of the color scale, useful for diverging colormaps.",
    )
    norm: Any | None = Field(
        default=None,
        description="Normalization for the colormap.",
    )
    add_outline: bool = Field(
        default=False,
        description="Add a thin border around groups of dots.",
    )
    outline_width: tuple[float, ...] = Field(
        default=(0.3, 0.05),
        description="Width of the outline as a fraction of the scatter dot size.",
    )
    outline_color: tuple[str, ...] = Field(
        default=("black", "white"),
        description="Colors for the outline: border color and gap color.",
    )
    ncols: int = Field(
        default=4,
        description="Number of panels per row.",
    )
    hspace: float = Field(
        default=0.25,
        description="Height of the space between multiple panels.",
    )
    wspace: float | None = Field(
        default=None,
        description="Width of the space between multiple panels.",
    )
    return_fig: bool | None = Field(
        default=None,
        description="Return the matplotlib figure.",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot; do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If `True` or a `str`, save the figure.",
    )
    ax: Any | None = Field(
        default=None,
        description="A matplotlib axes object.",
    )
    kwargs: dict[str, Any] | None = Field(
        default=None,
        description="Additional arguments passed to `matplotlib.pyplot.scatter()`.",
    )


class ScanpyPlUmapQueryParameters(BaseModel):
    """Parameters for querying the Scanpy `pl.umap` API."""

    method_name: str = Field(
        default="sc.pl.umap",
        description="The name of the method to call.",
    )
    question_uuid: str | None = Field(
        default=None,
        description="Unique identifier for the question.",
    )
    adata: str = Field(
        ...,
        description="Annotated data matrix.",
    )
    color: str | list[str] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )
    mask_obs: str | None = Field(
        default=None,
        description="Mask for observations.",
    )
    gene_symbols: str | None = Field(
        default=None,
        description="Column name in `.var` DataFrame that stores gene symbols.",
    )
    use_raw: bool | None = Field(
        default=None,
        description="Use `.raw` attribute of `adata` for coloring with gene expression.",
    )
    sort_order: bool = Field(
        default=True,
        description="Plot data points with higher values on top for continuous annotations.",
    )
    edges: bool = Field(
        default=False,
        description="Show edges.",
    )
    edges_width: float = Field(
        default=0.1,
        description="Width of edges.",
    )
    edges_color: str | list[float] | list[str] = Field(
        default="grey",
        description="Color of edges.",
    )
    neighbors_key: str | None = Field(
        default=None,
        description="Key for neighbors connectivities.",
    )
    arrows: bool = Field(
        default=False,
        description="Show arrows (deprecated in favor of `scvelo.pl.velocity_embedding`).",
    )
    arrows_kwds: dict[str, Any] | None = Field(
        default=None,
        description="Arguments passed to `quiver()`.",
    )
    groups: str | None = Field(
        default=None,
        description="Restrict to specific categories in categorical observation annotation.",
    )
    components: str | list[str] | None = Field(
        default=None,
        description="Components to plot, e.g., ['1,2', '2,3']. Use 'all' to plot all available components.",
    )
    dimensions: int | None = Field(
        default=None,
        description="Number of dimensions to plot.",
    )
    layer: str | None = Field(
        default=None,
        description="Name of the AnnData object layer to plot.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot ('2d' or '3d').",
    )
    scale_factor: float | None = Field(
        default=None,
        description="Scale factor for the plot.",
    )
    color_map: str | Any | None = Field(
        default=None,
        description="Color map for continuous variables.",
    )
    cmap: str | Any | None = Field(
        default=None,
        description="Alias for `color_map`.",
    )
    palette: str | list[str] | Any | None = Field(
        default=None,
        description="Colors for plotting categorical annotation groups.",
    )
    na_color: str | tuple[float, ...] = Field(
        default="lightgray",
        description="Color for null or masked values.",
    )
    na_in_legend: bool = Field(
        default=True,
        description="Include missing values in the legend.",
    )
    size: float | list[float] | None = Field(
        default=None,
        description="Point size. If `None`, computed as 120000 / n_cells.",
    )
    frameon: bool | None = Field(
        default=None,
        description="Draw a frame around the scatter plot.",
    )
    legend_fontsize: int | float | str | None = Field(
        default=None,
        description="Font size for legend.",
    )
    legend_fontweight: int | str = Field(
        default="bold",
        description="Font weight for legend.",
    )
    legend_loc: str = Field(
        default="right margin",
        description="Location of legend.",
    )
    legend_fontoutline: int | None = Field(
        default=None,
        description="Line width of the legend font outline in pt.",
    )
    colorbar_loc: str = Field(
        default="right",
        description="Location of the colorbar.",
    )
    vmax: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Upper limit of the color scale.",
    )
    vmin: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Lower limit of the color scale.",
    )
    vcenter: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="Center of the color scale, useful for diverging colormaps.",
    )
    norm: Any | None = Field(
        default=None,
        description="Normalization for the colormap.",
    )
    add_outline: bool = Field(
        default=False,
        description="Add a thin border around groups of dots.",
    )
    outline_width: tuple[float, ...] = Field(
        default=(0.3, 0.05),
        description="Width of the outline as a fraction of the scatter dot size.",
    )
    outline_color: tuple[str, ...] = Field(
        default=("black", "white"),
        description="Colors for the outline: border color and gap color.",
    )
    ncols: int = Field(
        default=4,
        description="Number of panels per row.",
    )
    hspace: float = Field(
        default=0.25,
        description="Height of the space between multiple panels.",
    )
    wspace: float | None = Field(
        default=None,
        description="Width of the space between multiple panels.",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot; do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If `True` or a `str`, save the figure.",
    )
    ax: Any | None = Field(
        default=None,
        description="A matplotlib axes object.",
    )
    return_fig: bool | None = Field(
        default=None,
        description="Return the matplotlib figure.",
    )
    marker: str = Field(
        default=".",
        description="Marker symbol.",
    )
    kwargs: dict[str, Any] | None = Field(
        default=None,
        description="Additional arguments passed to `matplotlib.pyplot.scatter()`.",
    )


class ScanpyPlDrawGraphQueryParameters(BaseModel):
    """Parameters for querying the Scanpy `pl.draw_graph` API."""

    method_name: str = Field(
        default="sc.pl.draw_graph",
        description="The name of the method to call.",
    )
    question_uuid: str | None = Field(
        default=None,
        description="Unique identifier for the question.",
    )
    adata: str = Field(
        ...,
        description="Annotated data matrix.",
    )
    color: str | list[str] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )
    gene_symbols: str | None = Field(
        default=None,
        description="Column name in `.var` DataFrame that stores gene symbols.",
    )
    use_raw: bool | None = Field(
        default=None,
        description="Use `.raw` attribute of `adata` for coloring with gene expression.",
    )
    sort_order: bool = Field(
        default=True,
        description=(
            "For continuous annotations used as color parameter, "
            "plot data points with higher values on top of others."
        ),
    )
    edges: bool = Field(
        default=False,
        description="Show edges.",
    )
    edges_width: float = Field(
        default=0.1,
        description="Width of edges.",
    )
    edges_color: str | list[float] | list[str] = Field(
        default="grey",
        description="Color of edges.",
    )
    neighbors_key: str | None = Field(
        default=None,
        description="Where to look for neighbors connectivities.",
    )
    arrows: bool = Field(
        default=False,
        description="Show arrows (deprecated in favor of `scvelo.pl.velocity_embedding`).",
    )
    arrows_kwds: dict[str, Any] | None = Field(
        default=None,
        description="Arguments passed to `quiver()`.",
    )
    groups: str | list[str] | None = Field(
        default=None,
        description="Restrict to a few categories in categorical observation annotation.",
    )
    components: str | list[str] | None = Field(
        default=None,
        description="For instance, ['1,2', '2,3']. To plot all available components use components='all'.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot.",
    )
    legend_loc: str = Field(
        default="right margin",
        description="Location of legend.",
    )
    legend_fontsize: int | float | str | None = Field(
        default=None,
        description="Numeric size in pt or string describing the size.",
    )
    legend_fontweight: int | str = Field(
        default="bold",
        description="Legend font weight.",
    )
    legend_fontoutline: int | None = Field(
        default=None,
        description="Line width of the legend font outline in pt.",
    )
    colorbar_loc: str | None = Field(
        default="right",
        description="Where to place the colorbar for continuous variables.",
    )
    size: float | list[float] | None = Field(
        default=None,
        description="Point size. If None, is automatically computed as 120000 / n_cells.",
    )
    color_map: str | Any | None = Field(
        default=None,
        description="Color map to use for continuous variables.",
    )
    palette: str | list[str] | Any | None = Field(
        default=None,
        description="Colors to use for plotting categorical annotation groups.",
    )
    na_color: str | tuple[float, ...] = Field(
        default="lightgray",
        description="Color to use for null or masked values.",
    )
    na_in_legend: bool = Field(
        default=True,
        description="If there are missing values, whether they get an entry in the legend.",
    )
    frameon: bool | None = Field(
        default=None,
        description="Draw a frame around the scatter plot.",
    )
    vmin: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the lower limit of the color scale.",
    )
    vmax: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the upper limit of the color scale.",
    )
    vcenter: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the center of the color scale.",
    )
    norm: Any | None = Field(
        default=None,
        description="Normalization for the colormap.",
    )
    add_outline: bool = Field(
        default=False,
        description="Add a thin border around groups of dots.",
    )
    outline_width: tuple[float, ...] = Field(
        default=(0.3, 0.05),
        description="Width of the outline as a fraction of the scatter dot size.",
    )
    outline_color: tuple[str, ...] = Field(
        default=("black", "white"),
        description="Colors for the outline: border color and gap color.",
    )
    ncols: int = Field(
        default=4,
        description="Number of panels per row.",
    )
    hspace: float = Field(
        default=0.25,
        description="Height of the space between multiple panels.",
    )
    wspace: float | None = Field(
        default=None,
        description="Width of the space between multiple panels.",
    )
    return_fig: bool | None = Field(
        default=None,
        description="Return the matplotlib figure.",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot; do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If `True` or a `str`, save the figure.",
    )
    ax: Any | None = Field(
        default=None,
        description="A matplotlib axes object.",
    )
    layout: str | None = Field(
        default=None,
        description="One of the `draw_graph()` layouts.",
    )
    kwargs: dict[str, Any] | None = Field(
        default=None,
        description="Additional arguments passed to `matplotlib.pyplot.scatter()`.",
    )


class ScanpyPlSpatialQueryParameters(BaseModel):
    """Parameters for querying the Scanpy `pl.spatial` API."""

    method_name: str = Field(
        default="sc.pl.spatial",
        description="The name of the method to call.",
    )
    question_uuid: str | None = Field(
        default=None,
        description="Unique identifier for the question.",
    )
    adata: str = Field(
        ...,
        description="Annotated data matrix.",
    )
    color: str | list[str] | None = Field(
        default=None,
        description="Keys for annotations of observations/cells or variables/genes.",
    )
    gene_symbols: str | None = Field(
        default=None,
        description="Column name in `.var` DataFrame that stores gene symbols.",
    )
    use_raw: bool | None = Field(
        default=None,
        description="Use `.raw` attribute of `adata` for coloring with gene expression.",
    )
    layer: str | None = Field(
        default=None,
        description="Name of the AnnData object layer to plot.",
    )
    library_id: str | None = Field(
        default=None,
        description="Library ID for Visium data, e.g., key in `adata.uns['spatial']`.",
    )
    img_key: str | None = Field(
        default=None,
        description=(
            "Key for image data, used to get `img` and `scale_factor` from "
            "'images' and 'scalefactors' entries for this library."
        ),
    )
    img: Any | None = Field(
        default=None,
        description="Image data to plot, overrides `img_key`.",
    )
    scale_factor: float | None = Field(
        default=None,
        description="Scaling factor used to map from coordinate space to pixel space.",
    )
    spot_size: float | None = Field(
        default=None,
        description="Diameter of spot (in coordinate space) for each point.",
    )
    crop_coord: tuple[int, ...] | None = Field(
        default=None,
        description="Coordinates to use for cropping the image (left, right, top, bottom).",
    )
    alpha_img: float = Field(
        default=1.0,
        description="Alpha value for image.",
    )
    bw: bool = Field(
        default=False,
        description="Plot image data in grayscale.",
    )
    sort_order: bool = Field(
        default=True,
        description=(
            "For continuous annotations used as color parameter, plot data points "
            "with higher values on top of others."
        ),
    )
    groups: str | list[str] | None = Field(
        default=None,
        description="Restrict to specific categories in categorical observation annotation.",
    )
    components: str | list[str] | None = Field(
        default=None,
        description="For example, ['1,2', '2,3']. To plot all available components, use 'all'.",
    )
    projection: str = Field(
        default="2d",
        description="Projection of plot.",
    )
    legend_loc: str = Field(
        default="right margin",
        description="Location of legend.",
    )
    legend_fontsize: int | float | str | None = Field(
        default=None,
        description="Numeric size in pt or string describing the size.",
    )
    legend_fontweight: int | str = Field(
        default="bold",
        description="Legend font weight.",
    )
    legend_fontoutline: int | None = Field(
        default=None,
        description="Line width of the legend font outline in pt.",
    )
    colorbar_loc: str | None = Field(
        default="right",
        description="Where to place the colorbar for continuous variables.",
    )
    size: float = Field(
        default=1.0,
        description="Point size. If None, automatically computed as 120000 / n_cells.",
    )
    color_map: str | Any | None = Field(
        default=None,
        description="Color map to use for continuous variables.",
    )
    palette: str | list[str] | Any | None = Field(
        default=None,
        description="Colors to use for plotting categorical annotation groups.",
    )
    na_color: str | tuple[float, ...] | None = Field(
        default=None,
        description="Color to use for null or masked values.",
    )
    na_in_legend: bool = Field(
        default=True,
        description="If there are missing values, whether they get an entry in the legend.",
    )
    frameon: bool | None = Field(
        default=None,
        description="Draw a frame around the scatter plot.",
    )
    vmin: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the lower limit of the color scale.",
    )
    vmax: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the upper limit of the color scale.",
    )
    vcenter: str | float | Any | list[str | float | Any] | None = Field(
        default=None,
        description="The value representing the center of the color scale.",
    )
    norm: Any | None = Field(
        default=None,
        description="Normalization for the colormap.",
    )
    add_outline: bool = Field(
        default=False,
        description="Add a thin border around groups of dots.",
    )
    outline_width: tuple[float, ...] = Field(
        default=(0.3, 0.05),
        description="Width of the outline as a fraction of the scatter dot size.",
    )
    outline_color: tuple[str, ...] = Field(
        default=("black", "white"),
        description="Colors for the outline: border color and gap color.",
    )
    ncols: int = Field(
        default=4,
        description="Number of panels per row.",
    )
    hspace: float = Field(
        default=0.25,
        description="Height of the space between multiple panels.",
    )
    wspace: float | None = Field(
        default=None,
        description="Width of the space between multiple panels.",
    )
    return_fig: bool | None = Field(
        default=None,
        description="Return the matplotlib figure.",
    )
    show: bool | None = Field(
        default=None,
        description="Show the plot; do not return axis.",
    )
    save: str | bool | None = Field(
        default=None,
        description="If `True` or a `str`, save the figure.",
    )
    ax: Any | None = Field(
        default=None,
        description="A matplotlib axes object.",
    )
    kwargs: dict[str, Any] | None = Field(
        default=None,
        description="Additional arguments passed to `matplotlib.pyplot.scatter()`.",
    )


class ScanpyPlQueryBuilder(BaseQueryBuilder):
    """A class for building a AnndataIO query object."""

    def create_runnable(
        self,
        query_parameters: list["BaseAPIModel"],
        conversation: "Conversation",
    ) -> Callable:
        """Create a runnable object for executing queries.

        Create runnable using the LangChain `create_structured_output_runnable`
        method.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """
        runnable = conversation.chat.bind_tools(query_parameters)
        return runnable | PydanticToolsParser(tools=query_parameters)

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list["BaseModel"]:
        """Generate a AnnDataIOQuery object.

        Generates the object based on the given question, prompt, and
        BioChatter conversation. Uses a Pydantic model to define the API fields.
        Creates a runnable that can be invoked on LLMs that are qualified to
        parameterise functions.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                ScanpyPlQuery.

        Returns:
        -------
            ScanpyPlQuery: the parameterised query object (Pydantic model)

        """
        tools = [
            ScanpyPlScatterQueryParameters,
            ScanpyPlPcaQueryParameters,
            ScanpyPlTsneQueryParameters,
            ScanpyPlUmapQueryParameters,
            ScanpyPlDrawGraphQueryParameters,
            ScanpyPlSpatialQueryParameters,
        ]
        runnable = self.create_runnable(conversation=conversation, query_parameters=tools)
        return runnable.invoke(question)



"""Module for generating anndata queries using LLM tools."""

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from langchain_core.output_parsers import PydanticToolsParser
from pydantic import BaseModel, Field

from biochatter.api_agent.base.agent_abc import BaseAPIModel, BaseQueryBuilder
from biochatter.llm_connect import Conversation

if TYPE_CHECKING:
    from biochatter.llm_connect import Conversation

ANNDATA_IO_QUERY_PROMPT = """
You are a world class algorithm, computational biologist with world leading knowledge
of the anndata package.
You are also an expert algorithm to return structured output formats.

You will be asked to provide code to answer a specific questions involving the anndata package.
NEVER return a code snippet or code itself, instead you have to return a structured output format.
You will have to create a structured output formats containing method:argument fields.

You will be asked to read in an anndata object from any anndata api supported format OR
to concatenate the anndata objects.
FOR THE MapAnnData, BE SURE TO ALWAYS USE THE variable of the anndata GIVEN IN THE INPUT, REPLACE IT IN THE method_name
Use the tools available:
"""


class ConcatenateAnnData(BaseAPIModel):
    """Concatenate AnnData objects along an axis."""

    method_name: str = Field(default="anndata.concat", description="NEVER CHANGE")
    adatas: list | dict = Field(
        ...,
        description=(
            "The objects to be concatenated. "
            "Either a list of AnnData objects or a mapping of keys to AnnData objects."
        ),
    )
    axis: str = Field(
        default="obs",
        description="Axis to concatenate along. Can be 'obs' (0) or 'var' (1). Default is 'obs'.",
    )
    join: str = Field(
        default="inner",
        description="How to align values when concatenating. Options: 'inner' or 'outer'. Default is 'inner'.",
    )
    merge: str | Callable | None = Field(
        default=None,
        description=(
            "How to merge elements not aligned to the concatenated axis. "
            "Strategies include 'same', 'unique', 'first', 'only', or a callable function."
        ),
    )
    uns_merge: str | Callable | None = Field(
        default=None,
        description="How to merge the .uns elements. Uses the same strategies as 'merge'.",
    )
    label: str | None = Field(
        default=None,
        description="Column in axis annotation (.obs or .var) to place batch information. Default is None.",
    )
    keys: list | None = Field(
        default=None,
        description=(
            "Names for each object being concatenated. "
            "Used for column values or appended to the index if 'index_unique' is not None. "
            "Default is None."
        ),
    )
    index_unique: str | None = Field(
        default=None,
        description="Delimiter for making the index unique. When None, original indices are kept.",
    )
    fill_value: Any | None = Field(
        default=None,
        description="Value used to fill missing indices when join='outer'. Default behavior depends on array type.",
    )
    pairwise: bool = Field(
        default=False,
        description="Include pairwise elements along the concatenated dimension. Default is False.",
    )


class MapAnnData(BaseAPIModel):
    """Apply mapping functions to elements of AnnData."""

    method_name: str = Field(
        default="anndata.obs|var['annotation_name'].map",
        description=(
            "ALWAYS ALWAYS ALWAYS REPLACE THE anndata BY THE ONE GIVEN BY THE INPUT"
            "Specifies the AnnData attribute and operation being performed. "
            "For example, 'obs.map' applies a mapping function or dictionary to the specified column in `adata.obs`. "
            "This must always include the AnnData component and the `.map` operation. "
            "Adapt the component (e.g., 'obs', 'var', etc.) to the specific use case."
        ),
    )
    dics: dict | None = Field(default=None, description="Dictionary to map over.")


class ReadH5AD(BaseAPIModel):
    """Read .h5ad-formatted hdf5 file."""

    method_name: str = Field(default="io.read_h5ad", description="NEVER CHANGE")
    filename: str = Field(default="dummy.h5ad", description="Path to the .h5ad file")
    backed: str | None = Field(
        default=None,
        description="Mode to access file: None, 'r' for read-only",
    )
    as_sparse: str | None = Field(
        default=None,
        description="Convert to sparse format: 'csr', 'csc', or None",
    )
    as_sparse_fmt: str | None = Field(
        default=None,
        description="Sparse format if converting, e.g., 'csr'",
    )
    index_unique: str | None = Field(
        default=None,
        description="Make index unique by appending suffix if needed",
    )


class ReadZarr(BaseAPIModel):
    """Read from a hierarchical Zarr array store."""

    method_name: str = Field(default="io.read_zarr", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.zarr",
        description="Path or URL to the Zarr store",
    )


class ReadCSV(BaseAPIModel):
    """Read .csv file."""

    method_name: str = Field(default="io.read_csv", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.csv",
        description="Path to the .csv file",
    )
    delimiter: str | None = Field(
        None,
        description="Delimiter used in the .csv file",
    )
    first_column_names: bool | None = Field(
        None,
        description="Whether the first column contains names",
    )


class ReadExcel(BaseAPIModel):
    """Read .xlsx (Excel) file."""

    method_name: str = Field(default="io.read_excel", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.xlsx",
        description="Path to the .xlsx file",
    )
    sheet: str | None = Field(None, description="Sheet name or index to read from")
    dtype: str | None = Field(
        None,
        description="Data type for the resulting dataframe",
    )


class ReadHDF(BaseAPIModel):
    """Read .h5 (hdf5) file."""

    method_name: str = Field(default="io.read_hdf", description="NEVER CHANGE")
    filename: str = Field(default="placeholder.h5", description="Path to the .h5 file")
    key: str | None = Field(None, description="Group key within the .h5 file")


class ReadLoom(BaseAPIModel):
    """Read .loom-formatted hdf5 file."""

    method_name: str = Field(default="io.read_loom", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.loom",
        description="Path to the .loom file",
    )
    sparse: bool | None = Field(None, description="Whether to read data as sparse")
    cleanup: bool | None = Field(None, description="Clean up invalid entries")
    X_name: str | None = Field(None, description="Name to use for X matrix")
    obs_names: str | None = Field(
        None,
        description="Column to use for observation names",
    )
    var_names: str | None = Field(
        None,
        description="Column to use for variable names",
    )


class ReadMTX(BaseAPIModel):
    """Read .mtx file."""

    method_name: str = Field(default="io.read_mtx", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.mtx",
        description="Path to the .mtx file",
    )
    dtype: str | None = Field(None, description="Data type for the matrix")


class ReadText(BaseAPIModel):
    """Read .txt, .tab, .data (text) file."""

    method_name: str = Field(default="io.read_text", description="NEVER CHANGE")
    filename: str = Field(
        default="placeholder.txt",
        description="Path to the text file",
    )
    delimiter: str | None = Field(None, description="Delimiter used in the file")
    first_column_names: bool | None = Field(
        None,
        description="Whether the first column contains names",
    )


class AnnDataIOQueryBuilder(BaseQueryBuilder):
    """A class for building a AnndataIO query object."""

    def create_runnable(
        self,
        query_parameters: list["BaseAPIModel"],
        conversation: "Conversation",
    ) -> Callable:
        """Create a runnable object for executing queries.

        Create runnable using the LangChain `create_structured_output_runnable`
        method.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """
        runnable = conversation.chat.bind_tools(query_parameters, tool_choice="required")
        return runnable | PydanticToolsParser(tools=query_parameters)

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list["BaseModel"]:
        """Generate a AnnDataIOQuery object.

        Generates the object based on the given question, prompt, and
        BioChatter conversation. Uses a Pydantic model to define the API fields.
        Creates a runnable that can be invoked on LLMs that are qualified to
        parameterise functions.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                AnnDataIOQuery.

        Returns:
        -------
            AnnDataIOQuery: the parameterised query object (Pydantic model)

        """
        tools = [
            ReadCSV,
            ReadExcel,
            ReadH5AD,
            ReadHDF,
            ReadLoom,
            ReadMTX,
            ReadText,
            ReadZarr,
            ConcatenateAnnData,
            MapAnnData,
        ]
        runnable = self.create_runnable(
            conversation=conversation,
            query_parameters=tools,
        )
        query = [
            ("system", ANNDATA_IO_QUERY_PROMPT),
            ("human", f"{question}"),
        ]
        return runnable.invoke(
            query,
        )



import json
import os
from collections.abc import Callable

import yaml

from ._misc import ensure_iterable, sentencecase_to_pascalcase
from .llm_connect import Conversation, LangChainConversation


class BioCypherPromptEngine:
    def __init__(
        self,
        schema_config_or_info_path: str | None = None,
        schema_config_or_info_dict: dict | None = None,
        model_provider: str = "google_genai",
        model_name: str = "gemini-2.0-flash",
        conversation_factory: Callable | None = None,
    ) -> None:
        """Given a biocypher schema configuration, extract the entities and
        relationships, and for each extract their mode of representation (node
        or edge), properties, and identifier namespace. Using these data, allow
        the generation of prompts for a large language model, informing it of
        the schema constituents and their properties, to enable the
        parameterisation of function calls to a knowledge graph.

        Args:
        ----
            schema_config_or_info_path: Path to a biocypher schema configuration
                file or the extended schema information output generated by
                BioCypher's `write_schema_info` function (preferred).

            schema_config_or_info_dict: A dictionary containing the schema
                configuration file or the extended schema information output
                generated by BioCypher's `write_schema_info` function
                (preferred).

            model_provider: The provider of the model to use.

            model_name: The name of the model to use for the conversation.
                DEPRECATED: This should now be set in the conversation factory.

            conversation_factory: A function used to create a conversation for
                creating the KG query. If not provided, a default function is
                used (creating an OpenAI conversation with the specified model,
                see `_get_conversation`).

        """
        if not schema_config_or_info_path and not schema_config_or_info_dict:
            raise ValueError(
                "Please provide the schema configuration or schema info as a path to a file or as a dictionary.",
            )

        if schema_config_or_info_path and schema_config_or_info_dict:
            raise ValueError(
                "Please provide the schema configuration or schema info as a "
                "path to a file or as a dictionary, not both.",
            )

        # set conversation factory or use default
        self.conversation_factory = conversation_factory if conversation_factory is not None else self._get_conversation

        if schema_config_or_info_path:
            # read the schema configuration
            with open(schema_config_or_info_path) as f:
                schema_config = yaml.safe_load(f)
        elif schema_config_or_info_dict:
            schema_config = schema_config_or_info_dict

        # check whether it is the original schema config or the output of
        # biocypher info
        is_schema_info = schema_config.get("is_schema_info", False)

        # extract the entities and relationships: each top level key that has
        # a 'represented_as' key
        self.entities = {}
        self.relationships = {}
        if not is_schema_info:
            for key, value in schema_config.items():
                # hacky, better with biocypher output
                name_indicates_relationship = "interaction" in key.lower() or "association" in key.lower()
                if "represented_as" in value:
                    if value["represented_as"] == "node" and not name_indicates_relationship:
                        self.entities[sentencecase_to_pascalcase(key)] = value
                    elif (value["represented_as"] == "node" and name_indicates_relationship) or value[
                        "represented_as"
                    ] == "edge":
                        self.relationships[sentencecase_to_pascalcase(key)] = value
        else:
            for key, value in schema_config.items():
                if not isinstance(value, dict):
                    continue
                if value.get("present_in_knowledge_graph", None) == False:
                    continue
                if value.get("is_relationship", None) == False:
                    self.entities[sentencecase_to_pascalcase(key)] = value
                elif value.get("is_relationship", None) == True:
                    value = self._capitalise_source_and_target(value)
                    self.relationships[sentencecase_to_pascalcase(key)] = value

        self.question = ""
        self.selected_entities = []
        self.selected_relationships = []  # used in property selection
        self.selected_relationship_labels = {}  # copy to deal with labels that
        # are not the same as the relationship name, used in query generation
        # dictionary to also include source and target types
        self.rel_directions = {}
        self.model_provider = model_provider
        self.model_name = model_name

    def _capitalise_source_and_target(self, relationship: dict) -> dict:
        """Make sources and targets PascalCase to match the entities. Sources and
        targets can be strings or lists of strings.
        """
        if "source" in relationship:
            if isinstance(relationship["source"], str):
                relationship["source"] = sentencecase_to_pascalcase(
                    relationship["source"],
                )
            elif isinstance(relationship["source"], list):
                relationship["source"] = [sentencecase_to_pascalcase(s) for s in relationship["source"]]
        if "target" in relationship:
            if isinstance(relationship["target"], str):
                relationship["target"] = sentencecase_to_pascalcase(
                    relationship["target"],
                )
            elif isinstance(relationship["target"], list):
                relationship["target"] = [sentencecase_to_pascalcase(t) for t in relationship["target"]]
        return relationship

    def _select_graph_entities_from_question(
        self,
        question: str,
        conversation: Conversation,
    ) -> str:
        conversation.reset()
        success1 = self._select_entities(
            question=question,
            conversation=conversation,
        )
        if not success1:
            raise ValueError(
                "Entity selection failed. Please try again with a different question.",
            )
        conversation.reset()
        success2 = self._select_relationships(conversation=conversation)
        if not success2:
            raise ValueError(
                "Relationship selection failed. Please try again with a different question.",
            )
        conversation.reset()
        success3 = self._select_properties(conversation=conversation)
        if not success3:
            raise ValueError(
                "Property selection failed. Please try again with a different question.",
            )

    def _generate_query_prompt(
        self,
        entities: list,
        relationships: dict,
        properties: dict,
        query_language: str | None = "Cypher",
    ) -> str:
        """Generate a prompt for a large language model to generate a database
        query based on the selected entities, relationships, and properties.

        Args:
        ----
            entities: A list of entities that are relevant to the question.

            relationships: A list of relationships that are relevant to the
                question.

            properties: A dictionary of properties that are relevant to the
                question.

            query_language: The language of the query to generate.

        Returns:
        -------
            A prompt for a large language model to generate a database query.

        """
        msg = (
            f"Generate a database query in {query_language} that answers "
            f"the user's question. "
            f"You can use the following entities: {entities}, "
            f"relationships: {list(relationships.keys())}, and "
            f"properties: {properties}. "
        )

        for relationship, values in relationships.items():
            self._expand_pairs(relationship, values)

        if self.rel_directions:
            msg += "Given the following valid combinations of source, relationship, and target: "
            for key, value in self.rel_directions.items():
                for pair in value:
                    msg += f"'(:{pair[0]})-(:{key})->(:{pair[1]})', "
            msg += f"generate a {query_language} query using one of these combinations. "

        msg += "Only return the query, without any additional text, symbols or characters --- just the query statement."
        return msg

    def generate_query_prompt(
        self,
        question: str,
        query_language: str | None = "Cypher",
    ) -> str:
        """Generate a prompt for a large language model to generate a database
        query based on the user's question and class attributes informing about
        the schema.

        Args:
        ----
            question: A user's question.

            query_language: The language of the query to generate.

        Returns:
        -------
            A prompt for a large language model to generate a database query.

        """
        self._select_graph_entities_from_question(
            question,
            self.conversation_factory(),
        )
        msg = self._generate_query_prompt(
            self.selected_entities,
            self.selected_relationship_labels,
            self.selected_properties,
            query_language,
        )
        return msg

    def generate_query(
        self,
        question: str,
        query_language: str | None = "Cypher",
    ) -> str:
        """Wrap entity and property selection and query generation; return the
        generated query.

        Args:
        ----
            question: A user's question.

            query_language: The language of the query to generate.

        Returns:
        -------
            A database query that could answer the user's question.

        """
        self._select_graph_entities_from_question(
            question,
            self.conversation_factory(),
        )

        return self._generate_query(
            question=question,
            entities=self.selected_entities,
            relationships=self.selected_relationship_labels,
            properties=self.selected_properties,
            query_language=query_language,
            conversation=self.conversation_factory(),
        )

    def _get_conversation(
        self,
        model_provider: str = "google_genai",
        model_name: str = "gemini-2.0-flash",
    ) -> "Conversation":
        """Create a conversation object given a model name.

        Args:
        ----
            model_provider: The provider of the model to use.

            model_name: The name of the model to use for the conversation.

        Returns:
        -------
            A BioChatter Conversation object for connecting to the LLM.

        Todo:
        ----
            Genericise to models outside of OpenAI.

        """
        conversation = LangChainConversation(
            model_provider=model_provider,
            model_name=model_name or self.model_name,
            prompts={},
            correct=False,
        )
        conversation.set_api_key(
            api_key=os.getenv("OPENAI_API_KEY"),
            user="test_user",
        )
        return conversation

    def _select_entities(
        self,
        question: str,
        conversation: "Conversation",
    ) -> bool:
        """Given a question, select the entities that are relevant to the question
        and store them in `selected_entities` and `selected_relationships`. Use
        LLM conversation to do this.

        Args:
        ----
            question: A user's question.

            conversation: A BioChatter Conversation object for connecting to the
                LLM.

        Returns:
        -------
            True if at least one entity was selected, False otherwise.

        """
        self.question = question

        conversation.append_system_message(
            "You have access to a knowledge graph that contains "
            f"these entity types: {', '.join(self.entities)}. Your task is "
            "to select the entity types that are relevant to the user's question "
            "for subsequent use in a query. Only return the entity types, "
            "comma-separated, without any additional text. Do not return "
            "entity names, relationships, or properties.",
        )

        msg, token_usage, correction = conversation.query(question)

        result = msg.split(",") if msg else []
        # TODO: do we go back and retry if no entities were selected? or ask for
        # a reason? offer visual selection of entities and relationships by the
        # user?

        if result:
            for entity in result:
                entity = entity.strip()
                if entity in self.entities:
                    self.selected_entities.append(entity)

        return bool(result)

    def _select_relationships(self, conversation: "Conversation") -> bool:
        """Given a question and the preselected entities, select relationships for
        the query.

        Args:
        ----
            conversation: A BioChatter Conversation object for connecting to the
                LLM.

        Returns:
        -------
            True if at least one relationship was selected, False otherwise.

        Todo:
        ----
            Now we have the problem that we discard all relationships that do
            not have a source and target, if at least one relationship has a
            source and target. At least communicate this all-or-nothing
            behaviour to the user.

        """
        if not self.question:
            raise ValueError(
                "No question found. Please make sure to run entity selection first.",
            )

        if not self.selected_entities:
            raise ValueError(
                "No entities found. Please run the entity selection step first.",
            )

        rels = {}
        source_and_target_present = False
        for key, value in self.relationships.items():
            if "source" in value and "target" in value:
                # if source or target is a list, expand to single pairs
                source = ensure_iterable(value["source"])
                target = ensure_iterable(value["target"])
                pairs = []
                for s in source:
                    for t in target:
                        pairs.append(
                            (
                                sentencecase_to_pascalcase(s),
                                sentencecase_to_pascalcase(t),
                            ),
                        )
                rels[key] = pairs
                source_and_target_present = True
            else:
                rels[key] = {}

        # prioritise relationships that have source and target, and discard
        # relationships that do not have both source and target, if at least one
        # relationship has both source and target. keep relationships that have
        # either source or target, if none of the relationships have both source
        # and target.

        if source_and_target_present:
            # First, separate the relationships into two groups: those with both
            # source and target in the selected entities, and those with either
            # source or target but not both.

            rels_with_both = {}
            rels_with_either = {}
            for key, value in rels.items():
                for pair in value:
                    if pair[0] in self.selected_entities:
                        if pair[1] in self.selected_entities:
                            rels_with_both[key] = value
                        else:
                            rels_with_either[key] = value
                    elif pair[1] in self.selected_entities:
                        rels_with_either[key] = value

            # If there are any relationships with both source and target,
            # discard the others.

            if rels_with_both:
                rels = rels_with_both
            else:
                rels = rels_with_either

            selected_rels = []
            for key, value in rels.items():
                if not value:
                    continue

                for pair in value:
                    if pair[0] in self.selected_entities or pair[1] in self.selected_entities:
                        selected_rels.append((key, pair))

            rels = json.dumps(selected_rels)
        else:
            rels = json.dumps(self.relationships)

        msg = (
            "You have access to a knowledge graph that contains "
            f"these entities: {', '.join(self.selected_entities)}. "
            "Your task is to select the relationships that are relevant "
            "to the user's question for subsequent use in a query. Only "
            "return the relationships without their sources or targets, "
            "comma-separated, and without any additional text. Here are the "
            "possible relationships and their source and target entities: "
            f"{rels}."
        )

        conversation.append_system_message(msg)

        res, token_usage, correction = conversation.query(self.question)

        result = res.split(",") if msg else []

        if result:
            for relationship in result:
                relationship = relationship.strip()
                if relationship in self.relationships:
                    self.selected_relationships.append(relationship)
                    rel_dict = self.relationships[relationship]
                    label = rel_dict.get("label_as_edge", relationship)
                    if "source" in rel_dict and "target" in rel_dict:
                        self.selected_relationship_labels[label] = {
                            "source": rel_dict["source"],
                            "target": rel_dict["target"],
                        }
                    else:
                        self.selected_relationship_labels[label] = {
                            "source": None,
                            "target": None,
                        }

        # if we selected relationships that have either source or target which
        # is not in the selected entities, we add those entities to the selected
        # entities.

        if self.selected_relationship_labels:
            for key, value in self.selected_relationship_labels.items():
                sources = ensure_iterable(value["source"])
                targets = ensure_iterable(value["target"])
                for source in sources:
                    if source is None:
                        continue
                    if source not in self.selected_entities:
                        self.selected_entities.append(
                            sentencecase_to_pascalcase(source),
                        )
                for target in targets:
                    if target is None:
                        continue
                    if target not in self.selected_entities:
                        self.selected_entities.append(
                            sentencecase_to_pascalcase(target),
                        )

        return bool(result)

    @staticmethod
    def _validate_json_str(json_str: str):
        json_str = json_str.strip()
        if json_str.startswith("```json"):
            json_str = json_str[7:]
        if json_str.endswith("```"):
            json_str = json_str[:-3]
        return json_str.strip()

    def _select_properties(self, conversation: "Conversation") -> bool:
        """Given a question (optionally provided, but in the standard use case
        reused from the entity selection step) and the selected entities, select
        the properties that are relevant to the question and store them in
        the dictionary `selected_properties`.

        Returns
        -------
            True if at least one property was selected, False otherwise.

        """
        if not self.question:
            raise ValueError(
                "No question found. Please make sure to run entity and relationship selection first.",
            )

        if not self.selected_entities and not self.selected_relationships:
            raise ValueError(
                "No entities or relationships provided, and none available "
                "from entity selection step. Please provide "
                "entities/relationships or run the entity selection "
                "(`select_entities()`) step first.",
            )

        e_props = {}
        for entity in self.selected_entities:
            if self.entities[entity].get("properties"):
                e_props[entity] = list(
                    self.entities[entity]["properties"].keys(),
                )

        r_props = {}
        for relationship in self.selected_relationships:
            if self.relationships[relationship].get("properties"):
                r_props[relationship] = list(
                    self.relationships[relationship]["properties"].keys(),
                )

        msg = (
            "You have access to a knowledge graph that contains entities and "
            "relationships. They have the following properties. Entities:"
            f"{e_props}, Relationships: {r_props}. "
            "Your task is to select the properties that are relevant to the "
            "user's question for subsequent use in a query. Only return the "
            "entities and relationships with their relevant properties in compact "
            "JSON format, without any additional text. Return the "
            "entities/relationships as top-level dictionary keys, and their "
            "properties as dictionary values. "
            "Do not return properties that are not relevant to the question."
        )

        conversation.append_system_message(msg)

        msg, token_usage, correction = conversation.query(self.question)
        msg = BioCypherPromptEngine._validate_json_str(msg)

        try:
            self.selected_properties = json.loads(msg) if msg else {}
        except json.decoder.JSONDecodeError:
            self.selected_properties = {}

        return bool(self.selected_properties)

    def _generate_query(
        self,
        question: str,
        entities: list,
        relationships: dict,
        properties: dict,
        query_language: str,
        conversation: "Conversation",
    ) -> str:
        """Generate a query in the specified query language that answers the user's
        question.

        Args:
        ----
            question: A user's question.

            entities: A list of entities that are relevant to the question.

            relationships: A list of relationships that are relevant to the
                question.

            properties: A dictionary of properties that are relevant to the
                question.

            query_language: The language of the query to generate.

            conversation: A BioChatter Conversation object for connecting to the
                LLM.

        Returns:
        -------
            A database query that could answer the user's question.

        """
        msg = self._generate_query_prompt(
            entities,
            relationships,
            properties,
            query_language,
        )

        conversation.append_system_message(msg)

        out_msg, token_usage, correction = conversation.query(question)

        return out_msg.strip()

    def _expand_pairs(self, relationship, values) -> None:
        if not self.rel_directions.get(relationship):
            self.rel_directions[relationship] = []
        if isinstance(values["source"], list):
            for source in values["source"]:
                if isinstance(values["target"], list):
                    for target in values["target"]:
                        self.rel_directions[relationship].append(
                            (source, target),
                        )
                else:
                    self.rel_directions[relationship].append(
                        (source, values["target"]),
                    )
        elif isinstance(values["target"], list):
            for target in values["target"]:
                self.rel_directions[relationship].append(
                    (values["source"], target),
                )
        else:
            self.rel_directions[relationship].append(
                (values["source"], values["target"]),
            )



"""Define the available models for the LLM connect module.

This module defines the available models for the LLM connect module and their
token limits.

The models are defined as an Enum, which allows for easy lookup of the model
token limits.
"""

from enum import Enum


class OpenAIModels(str, Enum):
    """Enum for OpenAI models."""

    GPT_35_TURBO = "gpt-3.5-turbo"
    GPT_35_TURBO_16K = "gpt-3.5-turbo-16k"
    GPT_35_TURBO_1106 = "gpt-3.5-turbo-1106"  # further updated 3.5-turbo
    GPT_4 = "gpt-4"
    GPT_4_32K = "gpt-4-32k"
    GPT_4_1106_PREVIEW = "gpt-4-1106-preview"  # gpt-4 turbo, 128k tokens
    GPT_4o = "gpt-4o"
    GPT_41 = "gpt-4.1"
    GPT_41_mini = "gpt-4.1-mini"


class GeminiModels(str, Enum):
    """Enum for Gemini models."""

    GEMINI_20_FLASH = "gemini-2.0-flash"
    GEMINI_25_FLASH = "gemini-2.5-flash-preview-04-17"


class MistralModels(str, Enum):
    """Enum for Mistral models."""

    MISTRAL_LARGE_LATEST = "mistral-large-latest"


class AnthropicModels(str, Enum):
    """Enum for Anthropic models."""

    CLAUDE_3_7_SONNET_LATEST = "claude-3-7-sonnet-latest"
    CLAUDE_3_5_HAIKU_LATEST = "claude-3-5-haiku-latest"


class HuggingFaceModels(str, Enum):
    """Enum for HuggingFace models."""

    BLOOM = "bigscience/bloom"


class XInferenceModels(str, Enum):
    """Enum for XInference models."""

    CUSTOM_ENDPOINT = "custom-endpoint"


class TokenLimits(Enum):
    """Enum for token limits of different models.

    Uses a tuple (model_type, token_limit) to ensure unique values while
    maintaining the token limit information.
    """

    GPT_35_TURBO = ("gpt-3.5-turbo", 4000)
    GPT_35_TURBO_16K = ("gpt-3.5-turbo-16k", 16000)
    GPT_35_TURBO_1106 = ("gpt-3.5-turbo-1106", 16000)
    GPT_4 = ("gpt-4", 8000)
    GPT_4_32K = ("gpt-4-32k", 32000)
    GPT_4_1106_PREVIEW = ("gpt-4-1106-preview", 128000)
    GPT_4o = ("gpt-4o", 32000)
    GPT_41 = ("gpt-4.1", 1048576)
    GPT_41_mini = ("gpt-4.1-mini", 1048576)
    BLOOM = ("bigscience/bloom", 1000)
    CUSTOM_ENDPOINT = ("custom-endpoint", 1)
    GEMINI_20_FLASH = ("gemini-2.0-flash", 1000000)
    GEMINI_25_FLASH = ("gemini-2.5-flash-preview-04-17", 1048576)

    @property
    def limit(self):
        """Return the token limit value."""
        return self.value[1]


# Define a list of models that support tool calling
TOOL_CALLING_MODELS = frozenset(
    [
        GeminiModels.GEMINI_20_FLASH.value,
        GeminiModels.GEMINI_25_FLASH.value,
        OpenAIModels.GPT_4o.value,
        OpenAIModels.GPT_41.value,
        OpenAIModels.GPT_41_mini.value,
        MistralModels.MISTRAL_LARGE_LATEST.value,
        AnthropicModels.CLAUDE_3_7_SONNET_LATEST.value,
        AnthropicModels.CLAUDE_3_5_HAIKU_LATEST.value,
    ]
)

# Define a list of models that support structured output
STRUCTURED_OUTPUT_MODELS = frozenset(
    [
        GeminiModels.GEMINI_20_FLASH.value,
        GeminiModels.GEMINI_25_FLASH.value,
        OpenAIModels.GPT_4o.value,
        OpenAIModels.GPT_41.value,
        OpenAIModels.GPT_41_mini.value,
    ]
)

# For backward compatibility (even if not sure if this is needed)
OPENAI_MODELS = [model.value for model in OpenAIModels]
GEMINI_MODELS = [model.value for model in GeminiModels]
HUGGINGFACE_MODELS = [model.value for model in HuggingFaceModels]
XINFERENCE_MODELS = [model.value for model in XInferenceModels]

# For backward compatibility and easy lookup
TOKEN_LIMITS = {
    OpenAIModels.GPT_35_TURBO.value: TokenLimits.GPT_35_TURBO.limit,
    OpenAIModels.GPT_35_TURBO_16K.value: TokenLimits.GPT_35_TURBO_16K.limit,
    OpenAIModels.GPT_35_TURBO_1106.value: TokenLimits.GPT_35_TURBO_1106.limit,
    OpenAIModels.GPT_4.value: TokenLimits.GPT_4.limit,
    OpenAIModels.GPT_4_32K.value: TokenLimits.GPT_4_32K.limit,
    OpenAIModels.GPT_4_1106_PREVIEW.value: TokenLimits.GPT_4_1106_PREVIEW.limit,
    HuggingFaceModels.BLOOM.value: TokenLimits.BLOOM.limit,
    XInferenceModels.CUSTOM_ENDPOINT.value: TokenLimits.CUSTOM_ENDPOINT.limit,
    GeminiModels.GEMINI_20_FLASH.value: TokenLimits.GEMINI_20_FLASH.limit,
    GeminiModels.GEMINI_25_FLASH.value: TokenLimits.GEMINI_25_FLASH.limit,
    OpenAIModels.GPT_41.value: TokenLimits.GPT_41.limit,
    OpenAIModels.GPT_41_mini.value: TokenLimits.GPT_41_mini.limit,
}



import json
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import Any, Literal

from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    ToolMessage,
)
from langchain_core.pydantic_v1 import ValidationError
from langgraph.graph import END, MessageGraph
from langgraph.graph.graph import CompiledGraph
from langsmith import traceable

logger = logging.getLogger(__name__)


class ReflexionAgentLogger:
    def __init__(self) -> None:
        self._logs: str = ""

    def log_step_message(
        self,
        step: int,
        node_name: str,
        output: BaseMessage,
    ):
        """Log step message
        Args:
          step int: step index
          output BaseMessage: step message
        """

    def log_final_result(self, final_result: dict[str, Any]) -> None:
        """Log final result
        Args:
          output BaseMessage: last step message
        """

    def _log_message(
        self,
        msg: str = "",
        level: Literal["info", "error", "warn"] | None = "info",
    ):
        """Save log message

        Args:
        ----
            msg: the message to be logged

            level: the log level to write

        """
        logger_func = logger.info if level == "info" else (logger.error if level == "error" else logger.warning)
        logger_func(msg)
        self._logs = self._logs + f"[{level}]" + f"{datetime.now().isoformat()} - {msg}\n"

    @property
    def logs(self):
        return self._logs


class ResponderWithRetries:
    """Raise request to LLM with 3 retries"""

    def __init__(self, runnable, validator):
        """Args:
        ----
        runnable: LLM agent
        validator: used to validate response

        """
        self.runnable = runnable
        self.validator = validator

    @traceable
    def respond(self, state: list[BaseMessage]):
        """Invoke LLM agent, this function will be called by LangGraph
        Args:
        state List[BaseMessage]: message history
        """
        response = []
        for attempt in range(3):
            try:
                response = self.runnable.invoke({"messages": state})
                self.validator.invoke(response)
                return response
            except ValidationError as e:
                state = state + [HumanMessage(content=repr(e))]
        return response


DRAFT_NODE = "draft"
EXECUTE_TOOL_NODE = "execute_tool"
REVISE_NODE = "revise"
END_NODE = END


class ReflexionAgentResult:
    def __init__(self, answer: str | None, tool_result: list[Any] | None):
        self.answer = answer
        self.tool_result = tool_result


class ReflexionAgent(ABC):
    r"""LLM agent reflexion framework:

    start -> draft -> execute tool -> revise -> evaluation -> end
                        /|\                        |
                         ---------------------------
    """

    RECURSION_LIMIT = 30

    def __init__(
        self,
        conversation_factory: Callable,
        max_steps: int | None = 20,
        agent_logger: ReflexionAgentLogger | None = ReflexionAgentLogger(),
    ):
        """Args:
        ----
          conversation_factory Callable: the callback to create Conversation
          max_steps int: max steps for reflextion loop

        """
        if max_steps <= 0:
            max_steps = ReflexionAgent.RECURSION_LIMIT
        recursion_limit = ReflexionAgent.RECURSION_LIMIT
        if recursion_limit < max_steps:
            recursion_limit = max_steps
        self.initial_responder = None
        self.revise_responder = None
        self.max_steps = max_steps
        self.recursion_limit = recursion_limit
        self.conversation = conversation_factory()
        self.agent_logger = agent_logger

    def _should_continue(self, state: list[BaseMessage]):
        """Determine if we need to continue reflexion
        Args:
          state List[BaseMessage]: message history
        """
        num_steps = ReflexionAgent._get_num_iterations(state)
        if num_steps > self.max_steps:
            return END
        return EXECUTE_TOOL_NODE

    @abstractmethod
    def _tool_function(self, state: list[BaseMessage]) -> ToolMessage:
        """Tool function, execute tool based on initial draft or revised answer
        Args:
          state List[BaseMessage]: message history
        Returns:
          ToolMessage
        """

    @abstractmethod
    def _create_initial_responder(
        self,
        prompt: str | None = None,
    ) -> ResponderWithRetries:
        """Draft responder, draft initial answer
        Args:
          prompt str: prompt for LLM to draft initial answer
        """

    @abstractmethod
    def _create_revise_responder(
        self,
        prompt: str | None = None,
    ) -> ResponderWithRetries:
        """Revise responder, revise answer according to tool function result
        Args:
          prompt str: prompt for LLM to draft initial answer
        """

    @abstractmethod
    def _parse_final_result(
        self,
        messages: list[BaseMessage],
    ) -> ReflexionAgentResult:
        """Parse the result of the last step
        Args:
          output BaseMessage: last step message
        Returns:
          ReflexionAgentResult: the parsed reuslt of the last step
        """

    def get_logs(self):
        return self.agent_logger.logs

    @staticmethod
    def _get_num_iterations(state: list[BaseMessage]):
        """Calculate iteration number
        Args:
          state List[BaseMessage]: message history

        Returns
        -------
          int: the iterations number

        """
        i = 0
        for m in state[::-1]:
            if not isinstance(m, (ToolMessage, AIMessage)):
                break
            i += 1
        return i

    @staticmethod
    def _get_user_question(state: list[BaseMessage]):
        """Get user's question from messages array"""
        for m in state:
            if not isinstance(m, HumanMessage):
                continue
            return m.content
        return None

    @staticmethod
    def _get_last_tool_result(messages: list[BaseMessage]):
        """Get result of the last tool node"""
        for m in messages[::-1]:
            if not isinstance(m, ToolMessage):
                continue
            content = json.loads(m.content)
            return content["result"]
        return None

    def _build_graph(self, prompt: str | None = None):
        """Build Langgraph graph for execution of chained LLM processes.

        Args:
        ----
          prompt str: prompt for LLM

        Returns:
        -------
          CompiledGraph | None: a Langgraph graph or None in case of errors

        """
        try:
            self.initial_responder = self._create_initial_responder(prompt)
            self.revise_responder = self._create_revise_responder(prompt)
            builder = MessageGraph()
            builder.add_node(DRAFT_NODE, self.initial_responder.respond)
            builder.add_node(EXECUTE_TOOL_NODE, self._tool_function)
            builder.add_node(REVISE_NODE, self.revise_responder.respond)
            builder.add_edge(DRAFT_NODE, EXECUTE_TOOL_NODE)
            builder.add_edge(EXECUTE_TOOL_NODE, REVISE_NODE)

            builder.add_conditional_edges(REVISE_NODE, self._should_continue)
            builder.set_entry_point(DRAFT_NODE)
            graph = builder.compile()
            return graph
        except Exception as e:
            logger.error(e)
            return None

    def _execute_graph(
        self,
        graph: CompiledGraph | None = None,
        question: str | None = "",
    ) -> ReflexionAgentResult:
        """Execute Langgraph graph
        Args:
          graph CompiledGraph: Langgraph graph
          question str: user question

        Returns
        -------
          answer str | None: string answer parsed from Langgraph graph execution

        """
        if graph is None:
            return None
        if len(question) == 0:
            return None

        events = graph.stream(
            [HumanMessage(content=question)],
            {
                "recursion_limit": self.recursion_limit,
            },
        )
        messages = [HumanMessage(content=question)]
        for i, step in enumerate(events):
            if isinstance(step, list):
                node, output = (f"{i}", step[i])
            else:
                node, output = next(iter(step.items()))
            self.agent_logger.log_step_message(i + 1, node, output)
            messages.append(output)

        final_result = self._parse_final_result(messages)
        self.agent_logger.log_final_result(final_result)
        return final_result

    def execute(
        self,
        question: str,
        prompt: str | None = None,
    ) -> ReflexionAgentResult:
        """Execute ReflexionAgent. Wrapper for building a graph and executing it,
        returning the final answer.

        Args:
        ----
          question str: user question
          prompt str: user prompt

        Returns:
        -------
          answer str | None: If it executes successfully, an answer to the
            question will be returned, otherwise, it returns None

        """
        if len(question) == 0:
            return None
        graph = self._build_graph(prompt)
        return self._execute_graph(graph, question)



"""OncoKB API agent."""

import uuid
from collections.abc import Callable
from typing import TYPE_CHECKING

import requests
from langchain.chains.openai_functions import create_structured_output_runnable
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.pydantic_v1 import BaseModel, Field

from biochatter.api_agent.base.agent_abc import (
    BaseFetcher,
    BaseInterpreter,
    BaseQueryBuilder,
)

if TYPE_CHECKING:
    from biochatter.llm_connect import Conversation


ONCOKB_QUERY_PROMPT = """
You are a world class algorithm for creating queries in structured formats. Your task is to use OncoKB Web APIs to
answer genomic questions.

For questions about genomic alterations, you can use the OncoKB API by providing the appropriate parameters based on the
type of query.

You have to extract the appropriate information out of the
Examples:
1. To annotate mutations by protein change, use the endpoint /annotate/mutations/byProteinChange with parameters like
    hugoSymbol, alteration, tumorType, etc.
2. To annotate copy number alterations, use the endpoint /annotate/copyNumberAlterations with parameters like
    hugoSymbol, copyNameAlterationType, tumorType, etc.

Use these formats to generate queries based on the question provided. Below is more information about the OncoKB API:
OncoKB API Documentation (Summary)

Base URL

https://demo.oncokb.org/api/v1

Endpoints and Parameters

1. Annotate Copy Number Alterations

	•	GET/POST /annotate/copyNumberAlterations
	•	Parameters:
	•	hugoSymbol: The gene symbol. Example: BRAF.
	•	entrezGeneId: The entrez gene ID. Example: 673.
	•	copyNameAlterationType: Copy number alteration type. Example: AMPLIFICATION.
	•	referenceGenome: Reference genome. Default: GRCh37.
	•	tumorType: Tumor type. Example: Melanoma.

2. Annotate Mutations by Genomic Change

	•	GET/POST /annotate/mutations/byGenomicChange
	•	Parameters:
	•	genomicLocation: Genomic location. Example: 7,140453136,140453136,A,T.
	•	referenceGenome: Reference genome. Default: GRCh37.
	•	tumorType: Tumor type. Example: Melanoma.

3. Annotate Mutations by HGVSg

	•	GET/POST /annotate/mutations/byHGVSg
	•	Parameters:
	•	hgvsg: HGVS genomic format. Example: 7:g.140453136A>T.
	•	referenceGenome: Reference genome. Default: GRCh37.
	•	tumorType: Tumor type. Example: Melanoma.

4. Annotate Mutations by Protein Change

	•	GET/POST /annotate/mutations/byProteinChange
	•	Parameters:
	•	hugoSymbol: The gene symbol. Example: BRAF.
	•	entrezGeneId: The entrez gene ID. Example: 673.
	•	alteration: Protein Change. Example: V600E.
	•	consequence: Consequence. Example: missense_variant.
	•	proteinStart: Protein Start. Example: 600.
	•	proteinEnd: Protein End. Example: 600.
	•	referenceGenome: Reference genome. Default: GRCh37.
	•	tumorType: Tumor type. Example: Melanoma.

5. Annotate Structural Variants

	•	GET/POST /annotate/structuralVariants
	•	Parameters:
	•	hugoSymbolA: Gene symbol A. Example: ABL1.
	•	entrezGeneIdA: Entrez gene ID A. Example: 25.
	•	hugoSymbolB: Gene symbol B. Example: BCR.
•	entrezGeneIdB: Entrez gene ID B. Example: 613.
	•	structuralVariantType: Structural variant type. Example: FUSION.
	•	isFunctionalFusion: Whether it is a functional fusion. Default: false.
	•	referenceGenome: Reference genome. Default: GRCh37.
	•	tumorType: Tumor type. Example: Melanoma.

6. Get Curated Genes

	•	GET /utils/allCuratedGenes
	•	Parameters:
	•	version: Data version.
	•	includeEvidence: Include gene summary and background.

7. Get Cancer Gene List

	•	GET /utils/cancerGeneList
	•	Parameters:
	•	version: Data version.
 """


ONCOKB_SUMMARY_PROMPT = """
You have to answer this question in a clear and concise manner: {question} Be factual!\n\
You are a world leading oncologist and molecular biologist who knows everything about OncoKB results.\n\
Do not make up information, only use the provided information and mention how relevant the found information is based on your knowledge about OncoKB\n\
Here is the information relevant to the question found on OncoKB:\n\
{context}
"""


class OncoKBQueryParameters(BaseModel):
    base_url: str = Field(
        default="https://demo.oncokb.org/api/v1",
        description="Base URL for the OncoKB API. Default is the demo instance.",
    )
    endpoint: str = Field(
        ...,
        description="Specific API endpoint to hit. Example: 'annotate/mutations/byProteinChange'.",
    )
    referenceGenome: str | None = Field(
        default="GRCh37",
        description="Reference genome, either GRCh37 or GRCh38. The default is GRCh37.",
    )
    hugoSymbol: str | None = Field(
        None,
        description="The gene symbol used in Human Genome Organisation. Example: BRAF.",
    )
    entrezGeneId: int | None = Field(
        None,
        description="The entrez gene ID. Higher priority than hugoSymbol. Example: 673.",
    )
    tumorType: str | None = Field(
        None,
        description="OncoTree(http://oncotree.info) tumor type name. The field supports OncoTree Code, OncoTree Name and OncoTree Main type. Example: Melanoma.",
    )
    alteration: str | None = Field(
        None,
        description="Protein Change. Example: V600E.",
    )
    consequence: str | None = Field(
        None,
        description="Consequence. Example: missense_variant.",
    )
    proteinStart: int | None = Field(
        None,
        description="Protein Start. Example: 600.",
    )
    proteinEnd: int | None = Field(
        None,
        description="Protein End. Example: 600.",
    )
    copyNameAlterationType: str | None = Field(
        None,
        description="Copy number alteration type. Available types: AMPLIFICATION, DELETION, GAIN, LOSS.",
    )
    structuralVariantType: str | None = Field(
        None,
        description="Structural variant type. Available values: DELETION, TRANSLOCATION, DUPLICATION, INSERTION, INVERSION, FUSION, UNKNOWN.",
    )
    isFunctionalFusion: bool | None = Field(
        default=False,
        description="Whether it is a functional fusion. Default value: false.",
    )
    hugoSymbolA: str | None = Field(
        None,
        description="The gene symbol A used in Human Genome Organisation. Example: ABL1.",
    )
    entrezGeneIdA: int | None = Field(
        None,
        description="The entrez gene ID A. Higher priority than hugoSymbolA. Example: 25.",
    )
    hugoSymbolB: str | None = Field(
        None,
        description="The gene symbol B used in Human Genome Organisation. Example: BCR.",
    )
    entrezGeneIdB: int | None = Field(
        None,
        description="The entrez gene ID B. Higher priority than hugoSymbolB. Example: 613.",
    )
    genomicLocation: str | None = Field(
        None,
        description="Genomic location. Example: 7,140453136,140453136,A,T.",
    )
    hgvsg: str | None = Field(
        None,
        description="HGVS genomic format. Example: 7:g.140453136A>T.",
    )
    question_uuid: str | None = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="Unique identifier for the question.",
    )


class OncoKBQueryBuilder(BaseQueryBuilder):
    """A class for building an OncoKBQuery object."""

    def create_runnable(
        self,
        query_parameters: "OncoKBQueryParameters",
        conversation: "Conversation",
    ) -> Callable:
        """Creates a runnable object for executing queries using the LangChain
        `create_structured_output_runnable` method.

        Args:
        ----
            query_parameters: A Pydantic data model that specifies the fields of
                the API that should be queried.

            conversation: A BioChatter conversation object.

        Returns:
        -------
            A Callable object that can execute the query.

        """
        return create_structured_output_runnable(
            output_schema=query_parameters,
            llm=conversation.chat,
            prompt=self.structured_output_prompt,
        )

    def parameterise_query(
        self,
        question: str,
        conversation: "Conversation",
    ) -> list[OncoKBQueryParameters]:
        """Generate an OncoKBQuery object.

        Generate based on the given question, prompt, and BioChatter
        conversation. Uses a Pydantic model to define the API fields. Creates a
        runnable that can be invoked on LLMs that are qualified to parameterise
        functions.

        Args:
        ----
            question (str): The question to be answered.

            conversation: The conversation object used for parameterising the
                OncoKBQuery.

        Returns:
        -------
            OncoKBQueryParameters: the parameterised query object (Pydantic model)

        """
        runnable = self.create_runnable(
            query_parameters=OncoKBQueryParameters,
            conversation=conversation,
        )
        oncokb_call_obj = runnable.invoke(
            {"input": f"Answer:\n{question} based on:\n {ONCOKB_QUERY_PROMPT}"},
        )
        oncokb_call_obj.question_uuid = str(uuid.uuid4())
        return [oncokb_call_obj]


class OncoKBFetcher(BaseFetcher):
    """A class for retrieving API results.

    Retrieve from OncoKB given a parameterized OncoKBQuery.
    """

    def __init__(self, api_token="demo"):
        self.headers = {
            "Authorization": f"Bearer {api_token}",
            "Accept": "application/json",
        }
        self.base_url = "https://demo.oncokb.org/api/v1"

    def fetch_results(
        self,
        request_data: list[OncoKBQueryParameters],
        retries: int | None = 3,
    ) -> str:
        """Submit the OncoKB query and fetch the results directly.

        No multi-step procedure, thus no wrapping of submission and retrieval in
        this case.

        Args:
        ----
            request_data: List of OncoKBQuery objects (Pydantic models)
                containing the OncoKB query parameters.

            retries: The number of retries to fetch the results.

        Returns:
        -------
            str: The results of the OncoKB query.

        """
        # For now, we only use the first query in the list
        query = request_data[0]

        # Submit the query and get the URL
        params = query.dict(exclude_unset=True)
        endpoint = params.pop("endpoint")
        params.pop("question_uuid")
        full_url = f"{self.base_url}/{endpoint}"
        response = requests.get(full_url, headers=self.headers, params=params)
        response.raise_for_status()

        # Fetch the results from the URL
        results_response = requests.get(response.url, headers=self.headers)
        results_response.raise_for_status()

        return results_response.text


class OncoKBInterpreter(BaseInterpreter):
    def summarise_results(
        self,
        question: str,
        conversation_factory: Callable,
        response_text: str,
    ) -> str:
        """Extract the answer from the BLAST results.

        Args:
        ----
            question (str): The question to be answered.
            conversation_factory: A BioChatter conversation object.
            response_text (str): The response.text returned by OncoKB.

        Returns:
        -------
            str: The extracted answer from the BLAST results.

        """
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    "You are a world class molecular biologist who knows "
                    "everything about OncoKB and cancer genomics. Your task is "
                    "to interpret results from OncoKB API calls and summarise "
                    "them for the user.",
                ),
                ("user", "{input}"),
            ],
        )
        summary_prompt = ONCOKB_SUMMARY_PROMPT.format(
            question=question,
            context=response_text,
        )
        output_parser = StrOutputParser()
        conversation = conversation_factory()
        chain = prompt | conversation.chat | output_parser
        answer = chain.invoke({"input": {summary_prompt}})
        return answer



from collections.abc import Collection
from typing import Literal

from pydantic import BaseModel, Field


class CalculateQCMetricsParams(BaseModel):
    adata: str = Field(..., description="Annotated data matrix")
    expr_type: str = Field("counts", description="Name of kind of values in X")
    var_type: str = Field("genes", description="The kind of thing the variables are")
    qc_vars: str = Field("", description="Keys for boolean columns of .var for control variables")
    percent_top: str = Field("50,100,200,500", description="Ranks for library complexity assessment")
    layer: str = Field(None, description="Layer to use for expression values")
    use_raw: bool = Field(False, description="Use adata.raw.X instead of adata.X")
    inplace: bool = Field(False, description="Place calculated metrics in adata's .obs and .var")
    log1p: bool = Field(True, description="Compute log1p transformed annotations")
    parallel: bool | None = Field(None, description="Parallel computation flag")

    class Config:
        arbitrary_types_allowed = True


class FilterCellsParams(BaseModel):
    data: str = Field(
        ...,
        description="The (annotated) data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    min_counts: int | None = Field(None, description="Minimum number of counts required for a cell to pass filtering.")
    min_genes: int | None = Field(
        None, description="Minimum number of genes expressed required for a cell to pass filtering."
    )
    max_counts: int | None = Field(None, description="Maximum number of counts required for a cell to pass filtering.")
    max_genes: int | None = Field(
        None, description="Maximum number of genes expressed required for a cell to pass filtering."
    )
    inplace: bool = Field(True, description="Perform computation inplace or return result.")
    copy: bool = Field(False, description="Whether to copy the data or modify it inplace.")

    class Config:
        arbitrary_types_allowed = True


class FilterGenesParams(BaseModel):
    data: str = Field(
        ...,
        description="An annotated data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    min_counts: int | None = Field(None, description="Minimum number of counts required for a gene to pass filtering.")
    min_cells: int | None = Field(
        None,
        description="Minimum number of cells in which the gene is expressed required for the gene to pass filtering.",
    )
    max_counts: int | None = Field(None, description="Maximum number of counts allowed for a gene to pass filtering.")
    max_cells: int | None = Field(
        None,
        description="Maximum number of cells in which the gene is expressed allowed for the gene to pass filtering.",
    )
    inplace: bool = Field(True, description="Perform computation inplace or return result.")
    copy: bool = Field(False, description="Whether to return a copy of the data (not modifying the original).")

    class Config:
        arbitrary_types_allowed = True


class HighlyVariableGenesParams(BaseModel):
    adata: str = Field(
        ..., description="Annotated data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes."
    )
    layer: str | None = Field(None, description="Use adata.layers[layer] for expression values instead of adata.X.")
    n_top_genes: int | None = Field(
        None, description="Number of highly-variable genes to keep. Mandatory if flavor='seurat_v3'."
    )
    min_mean: float = Field(
        0.0125,
        description="Minimum mean expression threshold for highly variable genes. Ignored if flavor='seurat_v3'.",
    )
    max_mean: float = Field(
        3, description="Maximum mean expression threshold for highly variable genes. Ignored if flavor='seurat_v3'."
    )
    min_disp: float = Field(
        0.5, description="Minimum dispersion threshold for highly variable genes. Ignored if flavor='seurat_v3'."
    )
    max_disp: float = Field(
        float("inf"),
        description="Maximum dispersion threshold for highly variable genes. Ignored if flavor='seurat_v3'.",
    )
    span: float = Field(
        0.3,
        description="Fraction of the data (cells) used in variance estimation for the loess model fit if flavor='seurat_v3'.",
    )
    n_bins: int = Field(
        20, description="Number of bins for binning the mean gene expression. Normalization is done per bin."
    )
    flavor: Literal["seurat", "cell_ranger", "seurat_v3", "seurat_v3_paper"] = Field(
        "seurat", description="The method to use for identifying highly variable genes."
    )
    subset: bool = Field(False, description="If True, subset to highly-variable genes, otherwise just indicate them.")
    inplace: bool = Field(True, description="Whether to place calculated metrics in .var or return them.")
    batch_key: str | None = Field(
        None, description="If specified, highly-variable genes are selected separately within each batch and merged."
    )
    check_values: bool = Field(
        True, description="Whether to check if counts in selected layer are integers (relevant for flavor='seurat_v3')."
    )

    class Config:
        arbitrary_types_allowed = True


class Log1pParams(BaseModel):
    data: str = Field(
        ...,
        description="The (annotated) data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    base: float | None = Field(None, description="Base of the logarithm. Natural logarithm is used by default.")
    copy: bool = Field(
        False, description="If True, a copy of the data is returned. Otherwise, the operation is done inplace."
    )
    chunked: bool | None = Field(
        None, description="Process the data matrix in chunks, which will save memory. Applies only to AnnData."
    )
    chunk_size: int | None = Field(None, description="Number of observations (n_obs) per chunk to process the data in.")
    layer: str | None = Field(None, description="Entry of layers to transform.")
    obsm: str | None = Field(None, description="Entry of obsm to transform.")

    class Config:
        arbitrary_types_allowed = True


class PCAParams(BaseModel):
    data: str = Field(
        ...,
        description="The (annotated) data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    n_comps: int | None = Field(
        None,
        description="Number of principal components to compute. Defaults to 50, or 1 - minimum dimension size of selected representation.",
    )
    layer: str | None = Field(None, description="If provided, which element of layers to use for PCA.")
    zero_center: bool = Field(
        True,
        description="If True, compute standard PCA from covariance matrix. If False, omit zero-centering variables.",
    )
    svd_solver: str | None = Field(
        None, description="SVD solver to use. Options: 'auto', 'arpack', 'randomized', 'lobpcg', or 'tsqr'."
    )
    random_state: int | None = Field(0, description="Change to use different initial states for the optimization.")
    return_info: bool = Field(
        False, description="Only relevant when not passing an AnnData. Whether to return PCA info."
    )
    mask_var: str | None = Field(
        None, description="To run PCA only on certain genes. Default is .var['highly_variable'] if available."
    )
    use_highly_variable: bool | None = Field(
        None,
        description="Whether to use highly variable genes only, stored in .var['highly_variable']. Deprecated in 1.10.0.",
    )
    dtype: str = Field("float32", description="Numpy data type string to which to convert the result.")
    chunked: bool = Field(
        False, description="If True, perform incremental PCA using sklearn IncrementalPCA or dask-ml IncrementalPCA."
    )
    chunk_size: int | None = Field(
        None, description="Number of observations to include in each chunk. Required if chunked=True."
    )
    copy: bool = Field(
        False,
        description="If True, a copy of the data is returned when AnnData is passed. Otherwise, the operation is done inplace.",
    )

    class Config:
        arbitrary_types_allowed = True


class NormalizeTotalParams(BaseModel):
    adata: str = Field(
        ...,
        description="The annotated data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    target_sum: float | None = Field(
        None,
        description="Target sum after normalization. If None, each cell will have total counts equal to the median before normalization.",
    )
    exclude_highly_expressed: bool = Field(
        False, description="If True, exclude highly expressed genes from normalization computation."
    )
    max_fraction: float = Field(
        0.05,
        description="If exclude_highly_expressed=True, consider a gene as highly expressed if it has more than max_fraction of the total counts in at least one cell.",
    )
    key_added: str | None = Field(
        None, description="Name of the field in adata.obs where the normalization factor is stored."
    )
    layer: str | None = Field(None, description="Layer to normalize instead of X. If None, normalize X.")
    inplace: bool = Field(
        True, description="Whether to update adata or return normalized copies of adata.X and adata.layers."
    )
    copy: bool = Field(
        False, description="Whether to modify the copied input object. Not compatible with inplace=False."
    )

    class Config:
        arbitrary_types_allowed = True


class RegressOutParams(BaseModel):
    adata: str = Field(..., description="The annotated data matrix.")
    keys: str | Collection[str] = Field(
        ...,
        description="Keys for observation annotation on which to regress on. Can be a single key or a collection of keys.",
    )
    layer: str | None = Field(None, description="Layer to regress on, if provided.")
    n_jobs: int | None = Field(
        None, description="Number of jobs for parallel computation. None means using default n_jobs."
    )
    copy: bool = Field(False, description="If True, a copy of the data will be returned. Otherwise, modifies in-place.")

    class Config:
        arbitrary_types_allowed = True


class ScaleParams(BaseModel):
    data: str = Field(
        ...,
        description="The (annotated) data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    zero_center: bool = Field(
        True, description="If False, omit zero-centering variables, which allows to handle sparse input efficiently."
    )
    max_value: float | None = Field(
        None, description="Clip (truncate) to this value after scaling. If None, do not clip."
    )
    copy: bool = Field(False, description="Whether this function should be performed inplace.")
    layer: str | None = Field(None, description="If provided, which element of layers to scale.")
    obsm: str | None = Field(None, description="If provided, which element of obsm to scale.")
    mask_obs: str | None = Field(
        None,
        description="Restrict the scaling to a certain set of observations. The mask is specified as a boolean array or a string referring to an array in obs.",
    )

    class Config:
        arbitrary_types_allowed = True


class SubsampleParams(BaseModel):
    data: str = Field(
        ...,
        description="The (annotated) data matrix of shape n_obs × n_vars. Rows correspond to cells and columns to genes.",
    )
    fraction: float | None = Field(None, description="Subsample to this fraction of the number of observations.")
    n_obs: int | None = Field(None, description="Subsample to this number of observations.")
    random_state: int | None = Field(0, description="Random seed to change subsampling.")
    copy: bool = Field(False, description="If an AnnData is passed, determines whether a copy is returned.")

    class Config:
        arbitrary_types_allowed = True


class DownsampleCountsParams(BaseModel):
    adata: str = Field(..., description="Annotated data matrix.")
    counts_per_cell: int | None = Field(
        None,
        description="Target total counts per cell. If a cell has more than ‘counts_per_cell’, it will be downsampled to this number. Can be an integer or integer ndarray with same length as number of observations.",
    )
    total_counts: int | None = Field(
        None,
        description="Target total counts. If the count matrix has more than total_counts, it will be downsampled to this number.",
    )
    random_state: int | None = Field(0, description="Random seed for subsampling.")
    replace: bool = Field(False, description="Whether to sample the counts with replacement.")
    copy: bool = Field(False, description="Determines whether a copy of adata is returned.")

    class Config:
        arbitrary_types_allowed = True


class RecipeZheng17Params(BaseModel):
    adata: str = Field(..., description="Annotated data matrix.")
    n_top_genes: int = Field(1000, description="Number of genes to keep.")
    log: bool = Field(True, description="Take logarithm. If True, log-transform data after filtering.")
    plot: bool = Field(False, description="Show a plot of the gene dispersion vs. mean relation.")
    copy: bool = Field(False, description="Return a copy of adata instead of updating it.")

    class Config:
        arbitrary_types_allowed = True


class RecipeWeinreb17Params(BaseModel):
    adata: str = Field(..., description="Annotated data matrix.")
    log: bool = Field(True, description="Logarithmize data? If True, log-transform the data.")
    mean_threshold: float = Field(0.01, description="Threshold for mean expression of genes.")
    cv_threshold: float = Field(2, description="Threshold for coefficient of variation (CV) for gene dispersion.")
    n_pcs: int = Field(50, description="Number of principal components to use.")
    svd_solver: str = Field("randomized", description="SVD solver to use.")
    random_state: int = Field(0, description="Random state for reproducibility of results.")
    copy: bool = Field(False, description="Return a copy if True, else modifies the original AnnData.")

    class Config:
        arbitrary_types_allowed = True


class RecipeSeuratParams(BaseModel):
    adata: str = Field(..., description="Annotated data matrix.")
    log: bool = Field(True, description="Logarithmize data? If True, log-transform the data.")
    plot: bool = Field(False, description="Show a plot of the gene dispersion vs. mean relation.")
    copy: bool = Field(False, description="Return a copy if True, else modifies the original AnnData.")

    class Config:
        arbitrary_types_allowed = True


class CombatParams(BaseModel):
    adata: str = Field(..., description="Annotated data matrix.")
    key: str = Field(
        "batch", description="Key to a categorical annotation from obs that will be used for batch effect removal."
    )
    covariates: list[str] | None = Field(
        None, description="Additional covariates such as adjustment variables or biological conditions."
    )
    inplace: bool = Field(True, description="Whether to replace adata.X or to return the corrected data.")

    class Config:
        arbitrary_types_allowed = True


class ScrubletParams(BaseModel):
    adata: str = Field(..., description="Annotated data matrix (n_obs × n_vars).")
    adata_sim: str | None = Field(
        None, description="Optional AnnData object from scrublet_simulate_doublets() with same number of vars as adata."
    )
    batch_key: str | None = Field(None, description="Optional obs column name discriminating between batches.")
    sim_doublet_ratio: float = Field(
        2.0, description="Number of doublets to simulate relative to the number of observed transcriptomes."
    )
    expected_doublet_rate: float = Field(0.05, description="Estimated doublet rate for the experiment.")
    stdev_doublet_rate: float = Field(0.02, description="Uncertainty in the expected doublet rate.")
    synthetic_doublet_umi_subsampling: float = Field(
        1.0, description="Rate for sampling UMIs when creating synthetic doublets."
    )
    knn_dist_metric: str = Field("euclidean", description="Distance metric used for nearest neighbor search.")
    normalize_variance: bool = Field(True, description="Normalize the data such that each gene has a variance of 1.")
    log_transform: bool = Field(False, description="Whether to log-transform the data prior to PCA.")
    mean_center: bool = Field(True, description="If True, center the data such that each gene has a mean of 0.")
    n_prin_comps: int = Field(
        30,
        description="Number of principal components used to embed the transcriptomes prior to KNN graph construction.",
    )
    use_approx_neighbors: bool = Field(
        False, description="Use approximate nearest neighbor method (annoy) for KNN classifier."
    )
    get_doublet_neighbor_parents: bool = Field(
        False, description="If True, return parent transcriptomes that generated the doublet neighbors."
    )
    n_neighbors: int | None = Field(None, description="Number of neighbors used to construct the KNN graph.")
    threshold: float | None = Field(None, description="Doublet score threshold for calling a transcriptome a doublet.")
    verbose: bool = Field(True, description="If True, log progress updates.")
    copy: bool = Field(False, description="If True, return a copy of adata with Scrublet results added.")
    random_state: int = Field(0, description="Initial state for doublet simulation and nearest neighbors.")

    class Config:
        arbitrary_types_allowed = True


class ScrubletSimulateDoubletsParams(BaseModel):
    adata: str = Field(
        ..., description="Annotated data matrix of shape n_obs × n_vars. Rows correspond to cells, columns to genes."
    )
    layer: str | None = Field(
        None, description="Layer of adata where raw values are stored, or 'X' if values are in .X."
    )
    sim_doublet_ratio: float = Field(
        2.0, description="Number of doublets to simulate relative to the number of observed transcriptomes."
    )
    synthetic_doublet_umi_subsampling: float = Field(
        1.0,
        description="Rate for sampling UMIs when creating synthetic doublets. If 1.0, simply add UMIs from two randomly sampled transcriptomes.",
    )
    random_seed: int = Field(0, description="Random seed for reproducibility.")

    class Config:
        arbitrary_types_allowed = True
